#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import inspect
import os.path
import time
import unittest
import re
import tempfile
from os.path import splitext
from os import mkdir, rmdir, remove

from stratuslab.Monitor import Monitor
from stratuslab.Registrar import Registrar
from stratuslab.Runner import Runner
from stratuslab.marketplace.Uploader import Uploader as marketplaceUploader
from stratuslab.Creator import Creator
from stratuslab.Exceptions import OneException, ClientException
from stratuslab.Exceptions import ConfigurationException
from stratuslab.Exceptions import ExecutionException
from stratuslab.Exceptions import InputException
from stratuslab.ConfigHolder import ConfigHolder
import Util
from stratuslab.marketplace.Downloader import Downloader
from stratuslab.marketplace.ManifestDownloader import ManifestDownloader
import stratuslab.ClusterTest as ClusterTest
import stratuslab.RegistrationTest as RegistrationTest
import stratuslab.LdapAuthenticationTest as LdapAuthenticationTest
from stratuslab.PersistentDisk import PersistentDisk
from stratuslab.Util import sleep, filePutContent
from stratuslab.Util import printStep, printInfo, printWarning, printError
from stratuslab.image.Image import Image
from stratuslab.ManifestInfo import ManifestInfo


VM_START_TIMEOUT = 5 * 60  # 5 min


class Testor(unittest.TestCase):
    configHolder = None
    testNames = []

    def setUp(self):
        self.vmIds = []
        self.image = 'BN1EEkPiBx87_uLj2-sdybSI-Xb'  # ttylinux v14 x86_64
        self.ubuntuImg = 'BXSSv_2udGkpKgi6fcCaVniz1Zd'
        self._unlinkFiles([self.sshKey, self.sshKeyPub])

    def tearDown(self):
        self._unlinkFiles([self.sshKey, self.sshKeyPub])

    def _unlinkFiles(self, filesList):
        for f in filesList:
            try:
                os.unlink(f)
            except:
                pass

    def __init__(self, methodName='dummy'):
        super(Testor, self).__init__(methodName)

        self.vmIds = []
        self.sshKey = '/tmp/id_rsa_smoke_test' + str(os.getpid())
        self.sshKeyPub = self.sshKey + '.pub'
        self.testsToRun = []
        self.quotaCpu = 2
        self.runner = None

        Testor.configHolder.assign(self)
        self._setFieldsFromEnvVars()

    def _setFieldsFromEnvVars(self):
        self._setSingleFieldFromEnvVar('requestedIpAddress', 'STRATUSLAB_REQUESTED_IP_ADDRESS')
        self._setSingleFieldFromEnvVar('p12Certificate', 'STRATUSLAB_P12_CERTIFICATE')
        self._setSingleFieldFromEnvVar('p12Password', 'STRATUSLAB_P12_PASSWORD')
        self._setSingleFieldFromEnvVar('pdiskEndpoint', 'STRATUSLAB_PDISK_ENDPOINT')
        self._exportEndpointIfNotInEnv()
        self._setSingleFieldFromEnvVar('endpoint', 'STRATUSLAB_ENDPOINT')

    def _setSingleFieldFromEnvVar(self, field, env):
        if env in os.environ:
            setattr(self, field, os.environ[env])

    def _exportEndpointIfNotInEnv(self):
        if Util.envEndpoint in os.environ:
            return
        if not self.frontendIp:
            raise ConfigurationException(
                'Missing environment variable %s or configuration parameter frontend_ip' % Util.envEndpoint)
        os.environ[Util.envEndpoint] = self.frontendIp

    def dummy(self):
        pass

    def runTests(self):
        suite = unittest.TestSuite()
        if self.testNames:
            tests = self.testNames
        else:
            tests = self._extractTestMethodNames()

        self._excludeTests(tests)

        for test in tests:
            suite.addTest(Testor(test))

        testResult = unittest.TextTestRunner(verbosity=2).run(suite)
        return testResult.wasSuccessful()

    def runMethod(self, method):
        return method()

    def runInstancePublicNetworkTest(self):
        """Start new instance, ping it via public network and ssh into it, then stop it."""
        self._runInstanceTest()

    def runInstanceLocalNetworkTest(self):
        """Start new instance, ping it via local network and ssh into it, then stop it."""
        self._runInstanceTest(withLocalNetwork=True,
                              cmdToRun=['ping -c 2 www.google.com'])

    def runInstanceRequestedNetworkTest(self):
        """Start new instance, ping it via requested IP address and ssh into it, then stop it."""

        self._checkAttributePresent(['requestedIpAddress'])
        if not self.requestedIpAddress:
            raise InputException('Missing definition for requested IP. Are you not missing --requested-ip-address?')
        runner = self._startVm(requestedIpAddress=self.requestedIpAddress)

        _, allocatedIp = runner.getNetworkDetail(runner.vmIds[0])

        self.assertEqual(self.requestedIpAddress, allocatedIp)

        self._repeatCall(self._ping, runner)
        self._repeatCall(self._loginViaSsh, runner, '/bin/true')
        self._stopVm(runner)

    def exceedCpuQuotaTest(self):
        """Start x instances, where x is the cpu quota +1, then stop them."""

        printInfo('Current cpu quota: %s, starting as many +1' % self.quotaCpu)
        try:
            self._startVm(instanceNumber=int(self.quotaCpu) + 1)
        except OneException, ex:
            message = "CPU quota exceeded (Quota: %s.0, Used: %s.0, Requested: 1.0)." % (self.quotaCpu, self.quotaCpu)
            self.assertTrue(message in str(ex), 'Quota not working, got %s expected %s' % (ex.message, message))
        else:
            self.fail('Quota not enforced')
        finally:
            self._createRunner().killInstances(self.runner.vmIds)

    def _excludeTests(self, tests):
        if self.testsToExclude:
            for test in self.testsToExclude.split(','):
                try:
                    tests.remove(test)
                except ValueError:
                    printWarning("WARNING: Test '%s' not in a list of defined tests." % test)

    def _runInstanceTest(self, withLocalNetwork=False, cmdToRun=['/bin/true'], msgRecipients=None):
        runner = self._startVm(withLocalNetwork=withLocalNetwork, msgRecipients=msgRecipients)
        self._repeatCall(self._ping, runner)
        for cmd in cmdToRun:
            self._repeatCall(self._loginViaSsh, runner, cmd)
        self._stopVm(runner)

    def _prepareLog(self, logFile):
        log = open(logFile, 'aw')
        log.write('\n' * 3 + '=' * 60 + '\n')
        log.write(str(datetime.datetime.now()) + '\n')
        log.write('=' * 60 + '\n' * 3)
        return log

    def _startVm(self, withLocalNetwork=False, requestedIpAddress=None,
                 instanceNumber=1, noCheckImageUrl=False, msgRecipients=None,
                 raiseOnFailed=True):
        self.runner = self._createRunner(withLocalNetwork, requestedIpAddress)
        self.runner.instanceNumber = instanceNumber

        self.runner.noCheckImageUrl = noCheckImageUrl

        if not msgRecipients:
            self.runner.msgRecipients = msgRecipients

        vmIds = self.runner.runInstance()
        self.vmIds.extend(vmIds)

        for vmId in vmIds:
            vmStarted = self.runner.waitUntilVmRunningOrTimeout(vmId, VM_START_TIMEOUT,
                                                                failOn='Failed')
            if not vmStarted and raiseOnFailed:
                error = 'Failed to start VM id: %s' % vmId
                Util.printError(error, exit=False)
                raise OneException(error)

        return self.runner

    def _createRunner(self, withLocalNetwork=False, requestedIpAddress=None,
                      persistentDiskUUID=None, image=None):
        Util.generateSshKeyPair(self.sshKey)

        if not image:
            image = self.image

        options = Runner.defaultRunOptions()
        options['username'] = self.testUsername
        options['password'] = self.testPassword
        options['userPublicKeyFile'] = self.sshKeyPub
        options['verboseLevel'] = self.verboseLevel
        options['specificAddressRequest'] = requestedIpAddress
        options['persistentDiskUUID'] = persistentDiskUUID
        options['pdiskEndpoint'] = self.pdiskEndpoint
        options['marketplaceEndpoint'] = self.marketplaceEndpoint

        if withLocalNetwork:
            options['isLocalIp'] = True

        configHolder = ConfigHolder(options)
        return Runner(image, configHolder)

    def _repeatCall(self, method, *args):
        numberOfRepetition = 60
        for _ in range(numberOfRepetition):
            failed = False
            try:
                if args:
                    method(*args)
                else:
                    method()
            except ExecutionException:
                failed = True
                time.sleep(10)
            else:
                break

        if failed:
            Util.printError('Failed executing method %s %s times, giving-up' % (method, numberOfRepetition), exit=False)
            raise

    def _ping(self, runner):

        for vmId in self.vmIds:
            _, ip = runner.getNetworkDetail(vmId)
            res = Util.ping(ip)
            if not res:
                raise ExecutionException('Failed to ping %s' % ip)

    def _loginViaSsh(self, runner, cmd):

        for vmId in self.vmIds:
            _, ip = runner.getNetworkDetail(vmId)
            res = Util.sshCmd(cmd, ip, sshKey=self.sshKey)
            if res:
                raise ExecutionException('Failed to SSH into machine for %s with return code %s' % (ip, res))

    def _stopVm(self, runner):
        runner.killInstances(self.vmIds)

    def _checkAttributePresentAndInitialized(self, attrs):
        self._checkAttributePresent(attrs)
        self._checkAttributeInitialized(attrs)

    def _checkAttributePresent(self, attrs):
        for attr in attrs:
            if attr not in self.__dict__:
                raise Exception('Missing attribute %s. Missing an option argument?' % attr)

    def _checkAttributeInitialized(self, attrs):
        for attr in attrs:
            if not getattr(self, attr):
                raise Exception('Attribute %s is not set. Missing an option argument?' % attr)

    def _generateDummyImage(self, filename, size=2):
        devNull = open(os.path.devnull, 'w')
        Util.execute(['dd', 'if=/dev/zero', 'of=%s' % filename, 'bs=1000000', 'count=%s' % size],
                     stdout=devNull, stderr=devNull)
        devNull.close()

    def registrarTest(self):
        """Register a new node with ONE server, check that it is properly registered and remove it"""
        configHolder = self.configHolder.copy()
        configHolder.options['infoDriver'] = 'kvm'
        configHolder.options['virtDriver'] = 'kvm'
        configHolder.options['transfertDriver'] = 'nfs'
        configHolder.options['networkDriver'] = 'dummy'
        configHolder.options['username'] = self.oneUsername
        configHolder.options['password'] = self.proxyOneadminPassword
        registrar = Registrar(configHolder)
        hostname = 'registrar.ip.test'
        vmId = registrar.register([hostname])
        monitor = Monitor(configHolder)
        info = monitor.nodeDetail([vmId])[0]
        self.assertEqual(hostname, info.name)
        registrar.deregister(hostname)
        self.assertRaises(Exception, monitor.nodeDetail, [vmId])


    def notificationTest(self):
        """Check notifications on VM state changes"""
        pass

        #notifier = NotificationUseCaseTest()

        #connection = None
        #channel = None

        #try:
        #    connection, channel = notifier.initializeMessageQueue()
        #    msgRecipients = notifier.createMsgRecipients()
        #    self._runInstanceTest(msgRecipients=msgRecipients)
        #    vmId = self.vmIds[0]
        #    notifier.checkNotificationMessages(connection, vmId)
        #finally:
        #    notifier.cleanUpMessageQueue(connection)

    def formatAvailableTests(self):
        result = "Available tests:\n"
        for testName, testDoc in self._extractTestDescriptions():
            result += "    - %s: %s\n" % (testName, testDoc)
        return result

    def _extractTestDescriptions(self):
        methods = []
        for attrib in self.__class__.__dict__:
            if self._isTestMethod(attrib):
                methods.append((attrib, self.__class__.__dict__[attrib].__doc__))
        return methods

    def _extractTestMethodNames(self):
        methods = []
        for attrib in self.__class__.__dict__:
            if self._isTestMethod(attrib):
                methods.append(attrib)
        return methods

    def _isTestMethod(self, attrib):
        return inspect.ismethod(getattr(self, attrib)) and \
               (attrib.lower().startswith('test') or attrib.lower().endswith('test')) and \
               not attrib.startswith('_')

    def createImageTest(self):
        """Create a machine image based on a given one."""

        self._checkAttributePresentAndInitialized(['imageIdCreateImage',
                                                   'authorEmailCreateImage'])
        self._doCreateImage()

    def _doCreateImage(self):
        remote_test_file = '$HOME/createImageTest-%s' % str(os.getpid())
        script = """#!/bin/sh
    touch %s
    """ % remote_test_file

        fd, script_file = tempfile.mkstemp('.sh', 'script')
        os.write(fd, script)
        os.close(fd)

        creator = self._createCreator(self.imageIdCreateImage,
                                      script_file=script_file)

        try:
            creator.create()
        except Exception, e:
            try:
                creator._stopMachine()
            except:
                pass
            raise e
        finally:
            os.unlink(script_file)

        timeout = 1000
        t_stop = time.time() + timeout
        t_step = 10
        printInfo("Waiting %i sec for image bundling." % timeout)

        while time.time() < t_stop:
            if creator.getVmState() in ('Done', 'Failed'):
                break
            printInfo('Time remaining: %d' % (t_stop - time.time()))
            time.sleep(t_step)

        # Assert instance state
        vm_state = creator.getVmState()
        if vm_state != 'Done':
            self.configHolder.username = self.testUsername
            self.configHolder.password = self.testPassword
            self.configHolder.endpoint = self.endpoint
            monitor = Monitor(self.configHolder)
            info = monitor._vmDetail(creator.getVmId())
            info_attributes = info.getAttributes()
            msg = "Image creation failed. Instance final state '%s'. Error: %s" % \
                  (vm_state, info_attributes.get('template_error_message',
                                                 'not set'))
            self.fail(msg)

        # Assert new image.
        # Assuming we are running on FE and can access VM log file.
        vm_id = creator.getVmId()
        mp_and_id = ''
        for line in open('/var/log/one/%s.log' % vm_id).readlines():
            if re.search('MARKETPLACE_AND_IMAGEID', line):
                mp_and_id = line.split()
                break
        image_id = mp_and_id[-1].strip('\'')
        if not Image.isImageId(image_id):
            self.fail("Failed to obtain new image ID.")
        marketplace_url = mp_and_id[-2].strip('\'')
        if not Util.isValidNetLocation(marketplace_url):
            self.fail("Failed to get marketplace endpoint.")

        self.configHolder.marketplaceEndpoint = marketplace_url
        ManifestDownloader(self.configHolder).getManifestInfo(image_id)

        self.image = image_id
        self.oneUsername = self.testUsername
        self.proxyOneadminPassword = self.testPassword
        cmds = ['python -c "import dirq"',
                'ls -l %s' % remote_test_file]

        self.marketplaceEndpoint = marketplace_url
        self._runInstanceTest(cmdToRun=cmds)

    def _createCreator(self, image, script_file=''):
        Util.generateSshKeyPair(self.sshKey)
        options = {}

        options['marketplaceEndpointNewimage'] = getattr(self, 'marketplaceEndpointUpload', '')

        options['authorEmail'] = self.authorEmailCreateImage
        options['saveDisk'] = True

        options['instanceType'] = 'c1.medium'

        options['verboseLevel'] = self.verboseLevel

        options['author'] = 'Jane Tester'
        options['comment'] = 'NB! Test image.'
        options['newImageGroupVersion'] = '1.99'
        options['newImageGroupName'] = 'base'
        options['newInstalledSoftwareName'] = 'Linux'
        options['newInstalledSoftwareVersion'] = '0.0'
        options['excludeFromCreatedImage'] = '/etc/resolve.conf,/usr/sbin/pppdump'
        options['scripts'] = script_file
        options['packages'] = 'python-dirq'
        # python-dirq is in Fedora base repo
        #options['extraOsReposUrls'] = 'http://download.fedora.redhat.com/pub/epel/5/i386/'

        options['installer'] = 'yum'
        options['os'] = 'linux'

        options['endpoint'] = self.endpoint
        options['username'] = self.testUsername
        options['password'] = self.testPassword

        options['userPublicKeyFile'] = self.sshKeyPub
        options['userPrivateKeyFile'] = self.sshKey

        options['shutdownVm'] = True

        options['marketplaceEndpoint'] = Downloader.ENDPOINT

        configHolder = ConfigHolder(options)

        return Creator(image, configHolder)

    def _registerInvalidImageInMarketplace(self):
        manifest_file = os.path.join(Util.getResourcesDir(),
                                     'manifest-invalid-sha1.xml')

        manifestInfo = ManifestInfo()
        manifestInfo.parseManifestFromFile(manifest_file)

        image_id = manifestInfo.identifier

        configHolder = ConfigHolder()
        configHolder.set('marketplaceEndpoint', self.marketplaceEndpoint)
        uploader = marketplaceUploader(configHolder)
        uploader.upload(manifest_file)

        return image_id

    def _startStopInvalidImage(self):
        mpendp_save = self.marketplaceEndpoint
        self.marketplaceEndpoint = self.marketplaceEndpointUpload or mpendp_save

        try:
            self.image = self._registerInvalidImageInMarketplace()
            info, vmId = self._startStopVmAndGetVmInfo()
        finally:
            self.marketplaceEndpoint = mpendp_save

        return info, vmId

    def oneReportsErrorViaXmlRpcTest(self):
        """Test if ONE reports error messages via XML RPC"""

        info, vmId = self._startStopInvalidImage()

        try:
            errorMessage = info.attribs['template_error_message']
        except KeyError:
            self.fail("No error message set.")
        else:
            self.failUnless(errorMessage, "Empty error message.")
            printInfo('VM %s failed with error message:\n%s' % (vmId, errorMessage))

    def _startStopVmAndGetVmInfo(self):
        """Return VM monitoring info and VM id."""

        self._startVm(noCheckImageUrl=True, raiseOnFailed=False)
        vmId = self.runner.vmIds[0]
        self._stopVm(self.runner)

        options = {}
        options['endpoint'] = getattr(self, 'endpoint')
        options['username'] = getattr(self, 'username', self.oneUsername)
        options['password'] = getattr(self, 'password', self.proxyOneadminPassword)

        monitor = Monitor(ConfigHolder(options))
        info = monitor._vmDetail(vmId)

        return info, vmId

    def marketPlaceTest(self):
        """Placeholder for marketplace test"""
        pass

    def clusterTest(self):
        """Cluster test"""
        ClusterTest.ClusterTest.sshKeyPub = self.sshKeyPub
        ClusterTest.ClusterTest.username = self.testUsername
        ClusterTest.ClusterTest.password = self.testPassword
        suite = self._createSuiteFromTestModule(ClusterTest)
        self._executeSuite(suite)

    def registrationTest(self):
        """Registration test"""
        suite = self._createSuiteFromTestModule(RegistrationTest)
        self._executeSuite(suite)

    def ldapAuthenticationTest(self):
        """LDAP authentication test"""
        suite = self._createSuiteFromTestModule(LdapAuthenticationTest)
        self._executeSuite(suite)

    def _createSuiteFromTestModule(self, module):
        suite = unittest.TestSuite()
        tests = unittest.TestLoader().loadTestsFromModule(module)
        suite.addTests(tests)
        return suite

    def _executeSuite(self, suite):
        testResult = unittest.TextTestRunner(verbosity=2).run(suite)
        self.assertTrue(testResult.wasSuccessful())

    def persistentDiskStorageTest(self):
        """Ensure that a disk can be created, written, stored and removed"""

        pdiskDevice = '/dev/hdc'  # !!!! Configured for the default image (ttylinux)
        pdiskMountPoint = '/mnt/pdisk-test'
        testFile = '%s/pdisk.txt' % pdiskMountPoint
        testFileCmp = '/tmp/pdisk.cmp'
        testString = 'pdiskTest'

        configHolder = Testor.configHolder.copy()
        configHolder.pdiskUsername = Testor.configHolder.testUsername
        configHolder.pdiskPassword = Testor.configHolder.testPassword
        pdisk = PersistentDisk(configHolder)

        Util.printAction('Creating a new persistent disk')
        diskUUID = pdisk.createVolume(1, 'test %s' % datetime.datetime.today(), False)

        Util.printAction('Checking persistent disk exists')
        if not pdisk.volumeExists(diskUUID):
            self.fail('An error occurred while creating a persistent disk')

        Util.printAction('Getting number of available users (before)')
        availableUserBeforeStart, _ = pdisk.getVolumeUsers(diskUUID)
        Util.printAction('Starting machine with persistent disk')
        runner = self._startVmWithPDiskAndWaitUntilUp(diskUUID)
        Util.printAction('Getting number of available users (after)')
        availableUserAfterStart, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterStart != (availableUserBeforeStart - 1):
            self.fail('Available users on persistent disk have to decrease by '
                      'one (%s, %s)' %
                      (availableUserBeforeStart, availableUserAfterStart))

        self._formatDisk(runner, pdiskDevice)
        self._mountDisk(runner, pdiskDevice, pdiskMountPoint)
        self._writeToFile(runner, testFile, testString)
        self._umountPDiskAndStopVm(runner, pdiskDevice)

        # Allow a few seconds for the disk to be dismounted.
        time.sleep(10)

        availableUserAfterStop, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterStop != availableUserBeforeStart:
            self.fail('Available users on persistent disk have to be the same '
                      'as when VM has started')

        runner = self._startVmWithPDiskAndWaitUntilUp(diskUUID)
        self._mountDisk(runner, pdiskDevice, pdiskMountPoint)
        self._writeToFile(runner, testFileCmp, testString)
        self._compareFiles(runner, testFile, testFileCmp)
        self._umountPDiskAndStopVm(runner, pdiskDevice)

        availableUserAfterStop, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterStop != availableUserBeforeStart:
            self.fail('Available users on persistent disk have to be the same '
                      'as when VM has started')

        Util.printAction('Removing persistent disk...')
        pdisk.deleteVolume(diskUUID)

        try:
            if pdisk.volumeExists(diskUUID):
                self.fail('The persistent disk %s is still present' % diskUUID)
        except ClientException, ex:
            if not re.match('404', ex.status):
                self.fail('The persistent disk %s is still present' % diskUUID)

    def persistentDiskStorageHotplugTest(self):
        """Ensure that a disk hot-plugged to a VM and then hot-unplugged"""

        pdiskDevice = '/dev/%s'
        pdiskMountPoint = '/mnt/pdisk-test'
        testFile = '%s/pdisk.txt' % pdiskMountPoint
        testFileCmp = '/tmp/pdisk.cmp'
        testString = 'pdiskTest'

        configHolder = Testor.configHolder.copy()
        configHolder.pdiskUsername = Testor.configHolder.testUsername
        configHolder.pdiskPassword = Testor.configHolder.testPassword
        pdisk = PersistentDisk(configHolder)

        runner = self._startVmWithPDiskAndWaitUntilUp(image=self.ubuntuImg)

        Util.printAction('Creating a new persistent disk')
        diskUUID = pdisk.createVolume(1, 'test %s' % datetime.datetime.today(),
                                      False)

        Util.printAction('Checking persistent disk exists')
        if not pdisk.volumeExists(diskUUID):
            self.fail('An error occurred while creating a persistent disk')

        self._modeprobe(runner, 'acpiphp')
        vmId = self.vmIds[0]
        node = runner.cloud.getVmNode(vmId)

        printStep('Attaching pdisk to VM')

        availableUserBeforeAttach, _ = pdisk.getVolumeUsers(diskUUID)
        device = pdisk.hotAttach(node, vmId, diskUUID)
        availableUserAfterAttach, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterAttach != (availableUserBeforeAttach - 1):
            self.fail('Available users on persistent disk have to decrease by '
                      'one; before=%s, after=%s' %
                      (availableUserBeforeAttach, availableUserAfterAttach))

        self._formatDisk(runner, pdiskDevice % device)
        self._mountDisk(runner, pdiskDevice % device, pdiskMountPoint)
        self._writeToFile(runner, testFile, testString)
        self._umountDisk(runner, pdiskDevice % device)

        printStep('Detaching pdisk of VM')
        pdisk.hotDetach(node, vmId, diskUUID)

        availableUserAfterDetach, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterDetach != availableUserBeforeAttach:
            self.fail('Available users on persistent disk have to be the '
                      'same as when VM has started; before=%s, after=%s' %
                      (availableUserBeforeAttach, availableUserAfterDetach))

        printStep('Re-attaching pdisk to VM')
        device = pdisk.hotAttach(node, vmId, diskUUID)

        self._mountDisk(runner, pdiskDevice % device, pdiskMountPoint)
        self._writeToFile(runner, testFileCmp, testString)
        self._compareFiles(runner, testFile, testFileCmp)
        self._umountPDiskAndStopVm(runner, pdiskDevice % device)

        availableUserAfterStop, _ = pdisk.getVolumeUsers(diskUUID)

        if availableUserAfterStop != availableUserBeforeAttach:
            self.fail('Available users on persistent disk have to be the '
                      'same as when VM has started; before=%s, after=%s' %
                      (availableUserBeforeAttach, availableUserAfterStop))

        Util.printAction('Removing persistent disk...')
        pdisk.deleteVolume(diskUUID)

        try:
            if pdisk.volumeExists(diskUUID):
                self.fail('The persistent disk %s is still present' % diskUUID)
        except ClientException, ex:
            if not re.match('404', ex.status):
                self.fail('The persistent disk %s is still present' % diskUUID)

    def persistentDiskStorageDownloadTest(self):
        """Check that an image can be downloaded correctly"""
        pdiskDevice = '/dev/hdc'  # !!!! Configured for the default image (ttylinux)
        pdiskMountPoint = '/mnt/pdisk-test'
        testFile = '%s/pdisk.txt'
        testString = 'pdiskTest'
        downloadedCompressedDisk = '/var/tmp/pdisk-img.gz'
        localMountPoint = '/mnt/pdisk-check'
        localTestFile = '/tmp/pdiskGzip.tmp'

        configHolder = Testor.configHolder.copy()
        configHolder.pdiskUsername = Testor.configHolder.testUsername
        configHolder.pdiskPassword = Testor.configHolder.testPassword
        pdisk = PersistentDisk(configHolder)

        Util.printAction('Creating a new persistent disk')
        diskUUID = pdisk.createVolume(1, 'test %s' % datetime.datetime.today(), False)

        Util.printAction('Checking persistent disk exists')
        if not pdisk.volumeExists(diskUUID):
            self.fail('An error occurred while creating a persistent disk')

        Util.printAction('Starting machine with persistent disk')
        runner = self._startVmWithPDiskAndWaitUntilUp(diskUUID)

        self._formatDisk(runner, pdiskDevice)
        self._mountDisk(runner, pdiskDevice, pdiskMountPoint)
        self._writeToFile(runner, testFile % pdiskMountPoint, testString)
        self._umountPDiskAndStopVm(runner, pdiskDevice)

        try:
            Util.printAction('Downloading volume...')
            # compressed disk comes in HTTP response - don't print it from HTTP client!
            verb_save = pdisk.client.verboseLevel
            pdisk.client.verboseLevel = 0
            pdisk.downloadVolume(diskUUID, downloadedCompressedDisk)
            pdisk.client.verboseLevel = verb_save
            volume = self._gunzip(downloadedCompressedDisk)
        finally:
            try:
                remove(downloadedCompressedDisk)
            except:
                pass

        try:
            if not self._localMount(volume, localMountPoint, ['loop', ]):
                self.fail('Error mounting downloaded image, corrupted?')

            filePutContent(localTestFile, testString)

            fileEquals = self._compareLocalFiles(localTestFile, testFile % localMountPoint)

            if not fileEquals:
                self.fail('Downloaded volume is corrupted')

            self._localUmount(localMountPoint)
        finally:
            Util.printAction('Post test clean-up...')
            try:
                remove(volume)
            except:
                pass
            try:
                rmdir(localMountPoint)
            except:
                pass

    def _startVmWithPDiskAndWaitUntilUp(self, pdisk=None, image=None):
        runner = self._createRunner(persistentDiskUUID=pdisk, image=image)
        vmIds = runner.runInstance()
        if len(vmIds) < 1:
            self.fail('An error occurred while starting a VM')
        if not runner.waitUntilVmRunningOrTimeout(vmIds[0], failOn='Failed'):
            self.fail('Failed starting VM: image %s, pdisk %s' % (str(image or runner.vm_image),
                                                                  str(pdisk)))
        self.vmIds.extend(vmIds)
        self._repeatCall(self._ping, runner)
        self._repeatCall(self._loginViaSsh, runner, '/bin/true')
        return runner

    def _umountPDiskAndStopVm(self, runner, pdiskDevice):
        self._umountDisk(runner, pdiskDevice)
        Util.printStep('Stopping VM...')
        self._stopVm(runner)
        self.vmIds = []
        # Wait for the pdisk hook to be executed
        sleep(20)

    def _formatDisk(self, runner, device):
        Util.printStep('Formating device %s' % device)
        mkfsScript = '/tmp/mkfs'
        self._loginViaSsh(runner, 'echo "echo \'y\' | /sbin/mkfs.ext3 %s" > %s'
                                  % (device, mkfsScript))
        self._loginViaSsh(runner, 'bash %s' % mkfsScript)

    def _mountDisk(self, runner, device, mountPoint):
        Util.printStep('Mounting device %s in %s...' % (device, mountPoint))
        self._loginViaSsh(runner, 'mkdir -p %s' % mountPoint)
        self._loginViaSsh(runner, 'mount %s %s' % (device, mountPoint))

    def _localMount(self, device, mountPoint, options=[]):
        Util.printStep('Mounting device %s on %s with options %s' % (device, mountPoint,
                                                                     (', '.join(options) or '<no options>')))
        mountOptions = ['-o', ','.join(options)]
        try:
            mkdir(mountPoint)
        except:
            pass
        rc, output = Util.executeGetStatusOutput(['mount'] + mountOptions + [device, mountPoint],
                                                 verboseLevel=self.verboseLevel)
        if output:
            printInfo(output)
        return rc == 0 and True or False

    def _localUmount(self, device):
        Util.printStep('Unmounting device %s...' % device)
        Util.execute(['umount', device])

    def _umountDisk(self, runner, device):
        Util.printStep('Unmounting device %s...' % device)
        self._loginViaSsh(runner, 'umount %s' % device)

    def _writeToFile(self, runner, filename, what):
        Util.printStep('Writing content to %s...' % filename)
        self._loginViaSsh(runner, 'echo "%s" > %s' % (what, filename))

    def _compareFiles(self, runner, file1, file2):
        Util.printStep('Comparing %s and %s content...' % (file1, file2))
        self._loginViaSsh(runner, 'diff %s %s' % (file1, file2))

    def _modeprobe(self, runner, module):
        Util.printStep('Loading module %s...' % module)
        self._loginViaSsh(runner, 'modprobe %s' % module)

    def _compareLocalFiles(self, file1, file2):
        Util.printStep('Comparing %s and %s content...' % (file1, file2))
        return Util.execute(['diff', file1, file2])

    def _gunzip(self, filename):
        Util.printStep("Unzipping file %s..." % filename)
        rc, output = Util.execute(['/bin/gunzip', filename], withOutput=True)
        if rc != 0:
            raise ExecutionException('Failed unzipping file: %s' % output)
        return splitext(filename)[0]



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import codecs
import os.path
import re
import subprocess
import sys
import time
import urllib2
import random
import urlparse
import hashlib
from random import sample
from string import ascii_lowercase
import platform

from stratuslab.Exceptions import ExecutionException, ValidationException
import stratuslab.api.LogUtil as LogUtil

import Defaults


# TODO: Move to Defaults
defaultRepoConfigSection = 'stratuslab_repo'
defaultRepoConfigPath = '.stratuslab/stratuslab.repo.cfg'
modulePath = os.path.abspath('%s/../' % os.path.abspath(os.path.dirname(__file__)))
systemsDir = '%s/stratuslab/system' % modulePath
varLibDir = '/var/lib/stratuslab/python'
defaultConfigFile = os.path.join(Defaults.ETC_DIR, 'stratuslab.cfg')
defaultConfigFileUser = os.path.join(Defaults.userHome, '.stratuslab', 'stratuslab-user.cfg')
manifestExt = '.xml'
cliLineSplitChar = '#'

VERBOSE_LEVEL_QUIET = 0
VERBOSE_LEVEL_NORMAL = 1
VERBOSE_LEVEL_DETAILED = 2

EMPHASIS_START = '\033[1;31m'
EMPHASIS_STOP = '\033[0m'

# Environment variable names
envEndpoint = 'STRATUSLAB_ENDPOINT'
userConfigFileSelectedSection = 'STRATUSLAB_USER_CONFIG_SECTION'

SSH_EXIT_STATUS_ERROR = 255
SSH_CONNECTION_RETRY_NUMBER = 2
SSH_CONNECTION_RETRY_SLEEP_MAX = 60


def getShareDir():
    # TODO: Windows compatibility?
    d = Defaults.SHARE_DIR
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../../../share')
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../resources/share')
    return d


def getTemplateDir():
    # TODO: Windows compatibility?
    d = Defaults.TEMPLATE_DIR
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../../../share/template')
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../resources/share/template')
    return d


def getResourcesDir():
    # TODO: Windows compatibility?
    d = Defaults.RESOURCES_DIR
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../../../share/resources')
    if not os.path.exists(d):
        d = os.path.join(os.path.dirname(__file__), '../../resources/share/resources')
    return d


def wget_as_xml(url, savePath):
    request = urllib2.Request(url, None, {'accept': 'application/xml'})
    fd = _wget(request)
    filePutContent(savePath, fd.read())
    fd.close()

def wget(url, savePath):
    fd = _wget(url)
    filePutContent(savePath, fd.read())
    fd.close()

def wstring_as_xml(url):
    request = urllib2.Request(url, None, {'accept': 'application/xml'})
    return wstring(request)

def wstring(url):
    fd = _wget(url)
    return fd.read()


def wread(url):
    return _wget(url)


def _wget(url):
    return urllib2.urlopen(url)


def ping(host, timeout=5, number=1, **kwargs):
    if systemName() == 'Darwin':
        timeout_opt = '-t'
    else:
        timeout_opt = '-w'
    p = subprocess.Popen(['ping', '-q', '-c', str(number),
                          timeout_opt, str(timeout), host], **kwargs)
    p.wait()
    success = (p.returncode == 0)
    return success


def appendOrReplaceInFile(filename, search, replace):
    if not os.path.isfile(filename):
        filePutContent(filename, replace)
        return

    fileContent = fileGetContent(filename)
    lines = fileContent.split('\n')

    newContent = []
    insertionMade = False
    for line in lines:
        if line.startswith(search):
            newContent.append(replace)
            insertionMade = True
        else:
            newContent.append(line)

    if insertionMade is False:
        newContent.append('%s\n' % replace)

    filePutContent(filename, '\n'.join(newContent))


def appendOrReplaceMultilineBlockInFile(filename, data, start='', until=''):
    content = fileGetContent(filename)
    newContent = appendOrReplaceMultilineBlockInString(content, data,
                                                       start=start, until=until)
    filePutContent(filename, newContent)


def appendOrReplaceMultilineBlockInString(content, data, start='', until=''):
    """
    A block in 'content' starts with the first line from 'data' or 'start'.
    Block ends right before 'until'.
    """
    if not until:
        until = '([ ]*|[ ]*#.*)$'

    if not start:
        data = data.strip()
        beginStr = data.split('\n', 1)[0]

        if not re.search(beginStr, content, re.M):
            return '%s%s%s\n\n' % (content, content.endswith('\n') and '\n' or '\n\n',
                                   data)
    else:
        beginStr = start

        beginStr_in_content = False
        for line in content.split('\n'):
            if line.startswith(beginStr):
                beginStr_in_content = True
                break

        if not beginStr_in_content:
            return '%s%s%s\n\n' % (content, content.endswith('\n') and '\n' or '\n\n',
                                   data)

    lines = content.split('\n')

    lineNums = []
    for i, line in enumerate(lines):
        if line.startswith(beginStr):
            lineNums.append(i)
            break

    insertIndex = lineNums[0]
    k = lineNums[0] + 1
    try:
        search_until = re.compile(until)
        while not search_until.match(lines[k]):
            lineNums.append(k)
            k += 1
    except IndexError:
        pass
    lineNums.reverse()
    for n in lineNums:
        del lines[n]
    if insertIndex == len(lines) - 1:
        data += '\n'
    lines.insert(insertIndex, data)

    return '\n'.join(lines)


def fileGetContent(filename):
    fd = open(filename, 'rb')
    content = fd.read()
    fd.close()
    return content


def filePutContent(filename, data, neverShowData=True):
    _printDetail('Creating file %s with content: \n%s\n' % (filename,
                                                            (neverShowData and '<hidden>' or data)))
    if isinstance(data, unicode):
        fh = codecs.open(filename, 'w', 'utf8')
    else:
        fh = open(filename, 'wb')
    fh.write(data)
    fh.close()


def fileAppendContent(filename, data):
    fd = open(filename, 'a')
    fd.write(data)
    fd.close()


def fileGetExtension(filename):
    try:
        ending = filename.rsplit('.', 1)[1]
    except IndexError:
        return ''
    if not ending:
        return ''
    return ending


def fileFind(directory, start='', end=''):
    try:
        for f in os.listdir(directory):
            if f.startswith(start) and f.endswith(end):
                return os.path.join(directory, f)
    except OSError:
        pass

    raise ValueError("Can't find file starting with %s and ending with %s in directory %s" % (start, end, directory))


def shaHexDigest(string):
    h = hashlib.sha1(string)
    return h.hexdigest()


def waitUntilPingOrTimeout(host, timeout, ticks=True, stdout=None, stderr=None):
    if not stdout:
        stdout = open(os.path.devnull, 'w')
    if not stderr:
        stderr = open(os.path.devnull, 'w')

    start = time.time()
    hostUp = False
    while not hostUp:
        if ticks:
            sys.stdout.flush()
            sys.stdout.write('.')
        hostUp = ping(host, stdout=stdout, stderr=stderr)
        sleep(1)

        if time.time() - start > timeout:
            if ticks:
                sys.stdout.flush()
                sys.stdout.write('\n')
            return False

    if ticks:
        sys.stdout.flush()
        sys.stdout.write('\n')
    return hostUp


def sleep(seconds):
    time.sleep(seconds)


def setPythonPath(path):
    if not path in sys.path:
        sys.path.append(path)


def execute(commandAndArgsList, **kwargs):
    wait = not kwargs.get('noWait', False)

    if 'noWait' in kwargs:
        del kwargs['noWait']

    output = kwargs.get('withOutput', False)
    if output:
        kwargs['stdout'] = subprocess.PIPE
        kwargs['stderr'] = subprocess.STDOUT
        kwargs['close_fds'] = True
    if 'withOutput' in kwargs:
        del kwargs['withOutput']

    if isinstance(commandAndArgsList, list):
        _cmd = ' '.join(commandAndArgsList)
    else:
        _cmd = commandAndArgsList

    _printDetail('Calling: %s' % _cmd, kwargs)

    if isinstance(commandAndArgsList, list) and kwargs.get('shell', False) is True:
        commandAndArgsList = ' '.join(commandAndArgsList)

    process = subprocess.Popen(commandAndArgsList, **kwargs)

    if wait:
        process.wait()

    if output:
        return process.returncode, process.stdout.read()
    else:
        return process.returncode


def executeGetStatusOutput(commandAndArgsList, **kwargs):
    kwargs['withOutput'] = True
    return execute(commandAndArgsList, **kwargs)


def executeRaiseOnError(cmd):
    res, output = execute(cmd.split(' '), withOutput=True)
    if res:
        raise ExecutionException('Failed executing %s with detail %s' % (cmd, output))


def _printDetail(message, kwargs={}):
    verboseLevel = _extractVerboseLevel(kwargs)
    verboseThreshold = _extractVerboseThreshold(kwargs)
    printDetail(message, verboseLevel, verboseThreshold)


def _extractVerboseLevel(kwargs):
    return _extractAndDeleteKey('verboseLevel', VERBOSE_LEVEL_QUIET, kwargs)


def _extractVerboseThreshold(kwargs):
    return _extractAndDeleteKey('verboseThreshold', VERBOSE_LEVEL_DETAILED, kwargs)


def _extractAndDeleteKey(key, default, dictionary):
    value = default
    if key in dictionary:
        value = dictionary[key]
        del dictionary[key]
    return value


def _format_action(msg):
    filler = ':' * (len(msg) + 6)
    fmt = '\n %s\n :: %s ::\n %s'
    return fmt % (filler, msg, filler)


def printAction(msg):
    LogUtil.info(_format_action(msg))


def printStep(msg):
    LogUtil.info(' :: %s' % msg)


def printInfo(msg):
    LogUtil.info(msg)


def printDebug(msg):
    LogUtil.debug(msg)


def printError(msg, exitCode=1, exit=True):
    # TODO: Revisit this design; having exit in API is not good.
    err = '  [ERROR] %s' % msg
    if exit:
        raise SystemExit(err)
    else:
        LogUtil.error(msg)


def printWarning(msg):
    LogUtil.warning('  [WARNING] %s' % msg)


def printDetail(msg, verboseLevel=1, verboseThreshold=1):
    # TODO: Review this to see if more levels are needed.
    if verboseThreshold == 0:
        LogUtil.info(msg)
    else:
        LogUtil.debug(msg)


def sshCmd(cmd, host, sshKey=None, port=22, user='root', timeout=5, passwordPrompts=0, **kwargs):
    def _appendToSshCommandFromKwargs(keyword, append):
        if kwargs.get(keyword, False):
            sshCmd.append(append)
        try:
            del kwargs[keyword]
        except:
            pass

    sshCmd = ['ssh',
              '-p', str(port),
              '-o', 'ConnectTimeout=%s' % timeout,
              '-o', 'StrictHostKeyChecking=no']
    sshCmd.append('-o NumberOfPasswordPrompts=%d' % int(passwordPrompts))

    if sshKey and os.path.isfile(sshKey):
        sshCmd.append('-i')
        sshCmd.append(sshKey)

    for keyAppend in [('sshVerb', '-v'), ('sshQuiet', '-q'), ('pseudoTTY', '-t')]:
        _appendToSshCommandFromKwargs(*keyAppend)

    sshCmd.append('%s@%s' % (user, host))
    sshCmd.append(cmd)

    output = None
    for i in range(0, SSH_CONNECTION_RETRY_NUMBER + 1):
        if i > 0:
            sleepTime = random.randint(0, SSH_CONNECTION_RETRY_SLEEP_MAX)
            _printDetail('[%i] Retrying ssh command in %i sec.' % (i, sleepTime), kwargs.copy())
            time.sleep(sleepTime)
        output = execute(sshCmd, **kwargs)
        if isinstance(output, int):
            es = output
        else:
            es = output[0]
        if es != SSH_EXIT_STATUS_ERROR:
            return output
    return output


def sshCmdWithOutput(cmd, host, sshKey=None, port=22, user='root', timeout=5, **kwargs):
    return sshCmd(cmd, host, sshKey=sshKey, port=port,
                  user=user, timeout=timeout, withOutput=True, **kwargs)


def sshCmdWithOutputVerb(cmd, host, sshKey=None, port=22, user='root', timeout=5, **kwargs):
    return sshCmd(cmd, host, sshKey=sshKey, port=port,
                  user=user, timeout=timeout, withOutput=True, sshVerb=True, **kwargs)


def sshCmdWithOutputQuiet(cmd, host, sshKey=None, port=22, user='root', timeout=5, **kwargs):
    return sshCmd(cmd, host, sshKey=sshKey, port=port,
                  user=user, timeout=timeout, withOutput=True, sshQuiet=True, **kwargs)


def sshInteractive(host, sshKey=None, port=22, user='root', timeout=5, passwordPrompts=3, **kwargs):
    return sshCmd('', host, sshKey=sshKey, port=port, user=user,
                  timeout=timeout, passwordPrompts=passwordPrompts, **kwargs)


def scp(src, dest, sshKey=None, port=22, **kwargs):
    scpCmd = ['scp', '-P', str(port), '-r', '-o', 'StrictHostKeyChecking=no']

    if sshKey and os.path.isfile(sshKey):
        scpCmd.append('-i')
        scpCmd.append(sshKey)

    scpCmd.append(src)
    scpCmd.append(dest)

    return execute(scpCmd, **kwargs)


def importSystem(system):
    try:
        return __import__(system)
    except ImportError, ex:
        msg = 'Error while importing module %s, with detail: ' % system
        printError(msg + str(ex), exit=False)
        raise


def validateIp(ipAddress):
    return isValidIpV4(ipAddress) or isValidIpV6(ipAddress)


def isValidIpV4(ip):
    """Validates IPv4 addresses.
    """
    pattern = re.compile(r"""
        ^
        (?:
          # Dotted variants:
          (?:
            # Decimal 1-255 (no leading 0's)
            [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
          |
            0x0*[0-9a-f]{1,2}  # Hexadecimal 0x0 - 0xFF (possible leading 0's)
          |
            0+[1-3]?[0-7]{0,2} # Octal 0 - 0377 (possible leading 0's)
          )
          (?:                  # Repeat 0-3 times, separated by a dot
            \.
            (?:
              [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
            |
              0x0*[0-9a-f]{1,2}
            |
              0+[1-3]?[0-7]{0,2}
            )
          ){0,3}
        |
          0x0*[0-9a-f]{1,8}    # Hexadecimal notation, 0x0 - 0xffffffff
        |
          0+[0-3]?[0-7]{0,10}  # Octal notation, 0 - 037777777777
        |
          # Decimal notation, 1-4294967295:
          429496729[0-5]|42949672[0-8]\d|4294967[01]\d\d|429496[0-6]\d{3}|
          42949[0-5]\d{4}|4294[0-8]\d{5}|429[0-3]\d{6}|42[0-8]\d{7}|
          4[01]\d{8}|[1-3]\d{0,9}|[4-9]\d{0,8}
        )
        $
    """, re.VERBOSE | re.IGNORECASE)
    return pattern.match(ip) is not None


def isValidIpV6(ip):
    """Validates IPv6 addresses.
    """
    pattern = re.compile(r"""
        ^
        \s*                         # Leading whitespace
        (?!.*::.*::)                # Only a single whildcard allowed
        (?:(?!:)|:(?=:))            # Colon iff it would be part of a wildcard
        (?:                         # Repeat 6 times:
            [0-9a-f]{0,4}           #   A group of at most four hexadecimal digits
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
        ){6}                        #
        (?:                         # Either
            [0-9a-f]{0,4}           #   Another group
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
            [0-9a-f]{0,4}           #   Last group
            (?: (?<=::)             #   Colon iff preceeded by exacly one colon
             |  (?<!:)              #
             |  (?<=:) (?<!::) :    #
             )                      # OR
         |                          #   A v4 address with NO leading zeros
            (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            (?: \.
                (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            ){3}
        )
        \s*                         # Trailing whitespace
        $
    """, re.VERBOSE | re.IGNORECASE | re.DOTALL)
    return pattern.match(ip) is not None


def unifyNetsize(netsize):
    classes = {'A': 2 ** 24, 'B': 2 ** 16, 'C': 2 ** 8}

    for letter, mask in classes.items():
        if netsize == letter:
            return mask

    return netsize


def networkSizeToNetmask(netsize):
    MAX_MASK_POW_TWO = 24
    MAX_MASK_LENGTH = 32
    for exponent in range(MAX_MASK_POW_TWO):
        if 2 ** exponent >= netsize:
            return MAX_MASK_LENGTH - exponent
    return MAX_MASK_LENGTH


def gatewayIpFromNetAddress(network):
    net, start = network.rsplit('.', 1)
    return '%s.%i' % (net, int(start) + 1)


def assignAttributes(instance, dictionary):
    for key, value in dictionary.items():
        setattr(instance, key, value)


def randomString(length=10):
    return ''.join(sample(list(ascii_lowercase), length))


def runMethodByName(obj, methodName, *args, **kw):
    return obj.__class__.__dict__[methodName](obj, *args, **kw)


def generateSshKeyPair(keyFilename):
    try:
        os.remove(keyFilename)
        os.remove(keyFilename + '.pub')
    except OSError:
        pass
    sshCmd = 'ssh-keygen -f %s -N "" -q' % keyFilename
    execute(sshCmd, shell=True)


def checkUrlExists(url, timeout=5):
    try:
        if not urllib2.urlopen(url, timeout=timeout):
            raise ExecutionException('urllib2.urlopen() did not return url handler.')
    except urllib2.URLError, ex:
        raise ValidationException(str(ex))
    return True


def constructEndPoint(fragment, protocol='https', port=8443, path=''):
    _protocol, _hostname, _port, _path = parseUri(fragment)

    _protocol = (_protocol and _protocol) or protocol + '://'
    _hostname = (_hostname and _hostname) or fragment
    _port = (_port and _port) or port
    _path = (_path and _path) or path

    return '%s%s:%s/%s' % (_protocol, _hostname, _port, _path)


def parseUri(uri):
    """Return tuple (protocol, hostname, port, path?query#fragment)"""
    m = re.match('([a-zA-Z0-9_]*://)?([^/:$]*):?(\d+)?/?(.*)', uri)
    protocol, hostname, port, path_and_query = m.group(1), m.group(2), m.group(3), m.group(4)
    return protocol, hostname, port, path_and_query


def sanitizeEndpoint(endpoint, protocol=Defaults.marketplaceProtocol, port=Defaults.marketplacePort):
    if not endpoint:
        return endpoint
    sanitized = endpoint
    _protocol, _, _, _ = parseUri(endpoint)
    if not _protocol:
        sanitized = constructEndPoint(sanitized, protocol, port)[:-1]  # trim trailing /
    return sanitized


def getHostnameFromUri(uri):
    return parseUri(uri)[1]


def getProtoFromUri(uri):
    return parseUri(uri)[0]


def getProtoHostnameFromUri(uri):
    return ''.join(parseUri(uri)[:2])


def getProtoHostnamePortFromUri(uri):
    groups = parseUri(uri)
    protoHost = ''.join(groups[:2])
    port = (groups[2] and ':%s' % groups[2]) or ''
    return protoHost + port


def getTimeInIso8601():
    """Return current time in iso8601 format."""
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(time.time()))


def toTimeInIso8601(_time):
    """Convert int or float to time in iso8601 format."""
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(_time))


def isTrueConfVal(var):
    return (str(var).lower() in ['true', 'yes', 'on', '1']) or False


def isFalseConfVal(var):
    return (str(var).lower() in ['false', 'no', 'off', '0', '']) or False


def importETree():
    try:
        import xml.etree.cElementTree as etree  # c-version first
        return etree
    except ImportError:
        try:
            import xml.etree.ElementTree as etree  # python version
            return etree
        except ImportError:
            raise Exception('failed to import ElementTree')


def escapeDoubleQuotes(string, times=1):
    return re.sub('"', '%s"' % ('\\' * times), string)


def sanitizePath(path):
    subs = [('\ ', ' ')]
    for s in subs:
        path = path.replace(*s)
    return path


def isValidNetLocation(url):
    r = urlparse.urlsplit(url)
    return (r.scheme and r.netloc) and True or False


def systemName():
    return platform.system()


def incrementMinorVersionNumber(version_string):
    vsplit = version_string.split('.')
    vsplit[1] = str(int(vsplit[1]) + 1)
    return '.'.join(vsplit)


def service(name, action):
    printDetail('Trying to %s %s' % (action, name))
    execute(['/etc/init.d/%s' % name, action])


def startService(name):
    service(name, 'start')


def stopService(name):
    service(name, 'stop')


def restartService(name):
    service(name, 'restart')


def getValueInKB(value):
    """Assume that if no unit specified, already in KB"""
    unit = ('KB', 'MB', 'GB')
    try:
        valueKB = int(value)
    except ValueError:
        valueNum = value[:-2]
        valueUnit = (value[-2:]).strip().upper()
        valueKB = int(valueNum) * (1024 ** unit.index(valueUnit))
    return str(valueKB)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from stratuslab.cloud import one

class CloudConnectorFactory:

    @staticmethod
    def getCloud(credentials, type='one'):
        if type == 'one':
            return one.OneConnector(credentials)
        else:
            raise NotImplementedError('Unknown cloud type: %s' % type)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import xmlrpclib

import urllib
from stratuslab.Exceptions import ValidationException
from stratuslab.Configurator import SimpleConfigParser
from stratuslab import Defaults

class AuthnFactory(object):
    
    @staticmethod
    def getCredentials(runnable):

        if AuthnFactory._useLocalHostCredentials(runnable):
            return LocalhostCredentialsConnector(runnable)
        
        try:
            usernamePasswordCredentials = runnable.username and runnable.password
        except AttributeError:
            usernamePasswordCredentials = False
        try:
            pemCredentials = runnable.pemCertificate and runnable.pemKey
        except AttributeError:
            pemCredentials = False
        
        if usernamePasswordCredentials:
            return UsernamePasswordCredentialsConnector(runnable)
        if pemCredentials:
            return CertificateCredentialsConnector(runnable)
        
        raise ValueError('Missing credentials')

    @staticmethod
    def _useLocalHostCredentials(runnable):
        endpoint = getattr(runnable, 'endpoint', None)
        frontend = getattr(runnable, 'frontendIp', None)

        if AuthnFactory._isLocalhost(endpoint):
            return True
        if AuthnFactory._isLocalhost(frontend):
            return True
        return False

    @staticmethod
    def _isLocalhost(host):
        return (host == 'localhost' or host == '127.0.0.1')

class CredentialsConnector(object):

    def __init__(self, runnable):
        self.runnable = runnable
        self.pathPrefix = ''

    def _manglePath(self, url):
        parts = url.split('/')
        return '/'.join(parts[0:-1]) + self.pathPrefix + parts[-1]

    def _insertUsernamePassword(self, url):
        protocolSeparator = '://'
        parts = url.split(protocolSeparator)
        quotedUsername = urllib.quote(self.username, '')
        quotedPassword = urllib.quote(self.password, '')
        return parts[0] + protocolSeparator + quotedUsername + ':' + quotedPassword + "@" + ''.join(parts[1:])
    
    def _validateUsernamePassword(self):
        if not self.username:
            raise ValidationException('Missing username')
        if not self.password:
            raise ValidationException('Missing password')

class UsernamePasswordCredentialsConnector(CredentialsConnector):
    
    def __init__(self, runnable):
        super(UsernamePasswordCredentialsConnector, self).__init__(runnable)
        self.username = runnable.username
        self.password = runnable.password
        self.pathPrefix = '/pswd/'
        self._validate()

    def _validate(self):
        return self._validateUsernamePassword()
    
    def createRpcConnection(self):
        url = self._insertUsernamePassword(self.runnable.cloud.server)
        url = self._manglePath(url)
        return xmlrpclib.ServerProxy(url)

    def createSessionString(self):
        return '%s:%s' % (self.username, self.password)


class CertificateCredentialsConnector(CredentialsConnector):
    
    class SafeTransportWithCert(xmlrpclib.SafeTransport):

        def __init__(self, cert, key):
            xmlrpclib.SafeTransport.__init__(self)
            self.__cert_file = cert
            self.__key_file = key

        def make_connection(self, host):
            host_with_cert = (host, {
                                     'key_file'  :  self.__key_file,
                                     'cert_file' :  self.__cert_file
                                     })
            return  xmlrpclib.SafeTransport.make_connection(self, host_with_cert)

    def __init__(self, runnable):
        super(CertificateCredentialsConnector, self).__init__(runnable)
        self.pemCertificate = runnable.pemCertificate
        self.pemKey = runnable.pemKey
        self.pathPrefix = '/cert/'
        self._validate()

    def _validate(self):
        if not os.path.exists(self.pemCertificate):
            raise ValueError('Can\'t find certificate file %s' % self.pemCertificate)
        
        if not os.path.exists(self.pemKey):
            raise ValueError('Can\'t find key file %s' % self.pemKey)
    
    def createRpcConnection(self):
        transport = CertificateCredentialsConnector.SafeTransportWithCert(self.pemCertificate, self.pemKey)
        url = self._manglePath(self.runnable.cloud.server)
        return xmlrpclib.ServerProxy(url, transport=transport)
    
    def createSessionString(self):
        return 'dummy:pass'


class LocalhostCredentialsConnector(CredentialsConnector):
    
    def __init__(self, runnable):
        super(LocalhostCredentialsConnector, self).__init__(runnable)
        self._setUsernamePassword(runnable)
        self._validate()
        
    def _validate(self):
        return self._validateUsernamePassword()
            
    def _setUsernamePassword(self, runnable):
        oneUsername = getattr(runnable, 'oneUsername', None)
        onePassword = getattr(runnable, 'onePassword', None)
        username = getattr(runnable, 'username', None)
        password = getattr(runnable, 'password', None)
        self.username = username or oneUsername
        self.password = password or onePassword
    
    def createRpcConnection(self):
        url = self._insertUsernamePassword('http://localhost:2633/RPC2')
        return xmlrpclib.ServerProxy(url)

    def createSessionString(self):
        return '%s:%s' % (self.username, self.password)


class UsernamePasswordCredentialsLoader(SimpleConfigParser):
    
    PASSWORD_INDEX = 0
    GROUP_INDEX = 1
     
    def __init__(self):
        super(UsernamePasswordCredentialsLoader, self).__init__()
        SimpleConfigParser.FILENAME = Defaults.AUTHN_CONFIG_FILE
        self.credentials = self.items

    def parse_value(self, value):
        parts = value.split(',')
        first = parts[0].strip()
        rest = ','.join(parts[1:]).strip()
        return (first, rest)
    
    def get_password(self, username):
        return self.credentials[username][UsernamePasswordCredentialsLoader.PASSWORD_INDEX]
            
    def get_group(self, username):
        return self.credentials[username][UsernamePasswordCredentialsLoader.GROUP_INDEX]



import base64
import json

from stratuslab.messaging.MsgClientFactory import getMsgClient
from stratuslab.Exceptions import InputException
from stratuslab.image.Image import Image

class SingleMessagePublisher(object):

    def __init__(self, message, configHolder):
        if not isinstance(message, basestring):
            raise InputException("Message should be 'str' or 'unicode'.")
        self.message = message
        self.client = getMsgClient(configHolder)

    def publish(self):
        self.client.deliver(self.message)

class ImageIdPublisher(SingleMessagePublisher):
    """Publishes base64 encoded JSON message with Marketplace image ID as 
    an entry in a map with a key 'imageid'.
    """

    IMAGEID_KEY = 'imageid'

    def __init__(self, message, image_id, configHolder):
        """message - empty string or JSON representation of a map 
        (can be base64 encoded).
        image_id - Marketplace image ID
        """
        super(ImageIdPublisher, self).__init__(message, configHolder)
        if not Image.isImageId(image_id):
            raise InputException('Marketplace image ID is expected.')
        self.message = self._set_imageid_on_message(self.message, image_id)

    def _set_imageid_on_message(self, message, imageid):
        'message - empty string or JSON and can be base64 encoded'
        if len(message) == 0:
            message = '{}'
        if not message.startswith('{'):
            # Assume this is base64 encoded message.
            message = base64.b64decode(message)
        try:
            message_dict = json.loads(message)
        except Exception, ex:
            raise ValueError("Couldn't load JSON message: %s" %  str(ex))
        message_dict[ImageIdPublisher.IMAGEID_KEY] = imageid
        return json.dumps(message_dict)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import urllib2

from Util import Util as MarketplaceUtil
from stratuslab.Exceptions import InputException, ExecutionException
from stratuslab.ManifestInfo import ManifestInfo
import stratuslab.Util as Util
from stratuslab import Defaults
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.ManifestInfo import _normalizeForInstanceAttribute as normalizeManifestElementForClassAttr

etree = Util.importETree()


class ManifestDownloader(object):
    ENDPOINT = Defaults.marketplaceEndpoint

    def __init__(self, configHolder=ConfigHolder()):
        self.configHolder = configHolder
        self.manifestObject = None
        self.marketplaceEndpoint = ManifestDownloader.ENDPOINT
        self.verboseLevel = 1

        configHolder.assign(self)

    def getManifestList(self, identifier):
        url = MarketplaceUtil.metadataUrl(self.marketplaceEndpoint, identifier)
        try:
            metadataEntries = Util.wstring_as_xml(url)
            return self._extractManifestInfos(self._parseXml(metadataEntries))
        except urllib2.HTTPError:
            raise InputException('Failed to find metadata entries: %s' % url)

    @staticmethod
    def _parseXml(xmlAsString):
        return etree.fromstring(xmlAsString)

    def _extractManifestInfos(self, manifestRootElement):
        manifestElements = manifestRootElement.findall('{http://www.w3.org/1999/02/22-rdf-syntax-ns#}RDF')
        manifests = []
        for e in manifestElements:
            manifest = ManifestInfo()
            manifest.parseManifestFromXmlTree(e)
            manifests.append(manifest)
        return manifests

    def getManifestInfo(self, resourceUri):
        """Return manifest as ManifestInfo object."""
        return self._getManifest(resourceUri)

    def getManifestAsFile(self, uri, filename):
        url = MarketplaceUtil.metadataUrl(self.marketplaceEndpoint, uri)
        self._downloadAsFile(url, filename)

    def _getManifest(self, resourceUri):
        url = MarketplaceUtil.metadataUrl(self.marketplaceEndpoint, resourceUri)
        self._printDetail('Downloading from: %s' % url)
        try:
            return self.__getManifest(url)
        except:
            Util.printError('Failed to get manifest for resource uri: %s' % resourceUri)
            raise

    def __getManifest(self, url):
        errorMessage = 'Failed to find metadata entry: %s' % url
        try:
            manifest = self._download(url)
        except urllib2.HTTPError:
            raise InputException(errorMessage)

        if manifest is not None:
            manifestInfo = ManifestInfo(self.configHolder)
        else:
            raise InputException(errorMessage)

        try:
            manifestInfo.parseManifestFromXmlTree(manifest)
        except ExecutionException, ex:
            raise InputException('Error parsing the metadata corresponding to url %s, with detail %s' % (url, ex))
        return manifestInfo

    def _download(self, url):
        try:
            xml = Util.wstring_as_xml(url)
            return self._parseXml(xml)
        except urllib2.URLError, ex:
            raise InputException('Failed to download: %s, with detail: %s' % (url, str(ex)))

    def _downloadAsFile(self, url, filename):
        try:
            return Util.wget_as_xml(url, filename)
        except urllib2.URLError, ex:
            raise InputException('Failed to download: %s, with detail: %s' % (url, str(ex)))

    def getImageLocations(self, imageId=''):
        return self.getImageElementValue('locations', imageId)

    def getImageVersion(self, imageId=''):
        return self.getImageElementValue('version', imageId)

    def getImageElementValue(self, element, imageId=''):
        if imageId:
            self._checkManifestAndImageId(imageId)
        elementNorm = normalizeManifestElementForClassAttr(element)
        try:
            return getattr(self.manifestObject, elementNorm)
        except AttributeError, ex:
            raise ExecutionException("Couldn't get '%s' element (normalized to '%s') from manifest. %s" %
                                     (element, elementNorm, str(ex)))

    def _checkManifestAndImageId(self, imageId):
        if not self.manifestObject:
            self.downloadManifestByImageId(imageId)
        if imageId.split('/')[0] != self.manifestObject.identifier:
            raise InputException('Given image ID [%s] does not match to downloaded [%s]' %
                                 (imageId, self.manifestObject.identifier))

    def downloadManifestByImageId(self, imageId):
        self.manifestObject = self._getManifest(imageId)

    def _printDetail(self, message):
        Util.printDetail(message, self.verboseLevel, 1)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import json
import stat
import base64
import mimetools
import mimetypes
import httplib2
import ssl
from httplib2 import httplib
from time import gmtime, strftime
from stratuslab import Util
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Exceptions import ServerException
from stratuslab.Exceptions import ClientException
from stratuslab.Exceptions import NetworkException

class HttpClient(object):

    def __init__(self, configHolder=ConfigHolder()):
        self.verboseLevel = None
        self.configHolder = configHolder
        self.crendentials = {}
        self.certificates = {}
        self.handleResponse = True
        self.useHttpCache = True
        configHolder.assign(self)

    def get(self, url, accept='application/xml'):
        return self._httpCall(url, 'GET', accept=accept)

    def post(self, url, body=None, contentType='application/xml', accept='application/xml'):
        return self._httpCall(url, 'POST', body, contentType, accept, retry=False)

    def post_multipart(self, url, files=[], params=[], accept='application/xml'):
        boundary, body = self._multipart_encode(files, params)
        contentType = 'multipart/form-data; boundary=%s' % boundary
        return self.post(url, body, contentType=contentType, accept=accept)

    def _multipart_encode(self, files, params):
        "files - list of (<attribute name>, <file descriptor>) tuples"
        "params - list of (<attribute name>, <value>) tuples"
        boundary = mimetools.choose_boundary()
        body = ''
        for(key, value) in params:
            body += '--%s\r\n' % boundary
            body += 'Content-Disposition: form-data; name="%s"' % key
            body += '\r\n\r\n' + value + '\r\n'
        for(key, fh) in files:
            file_size = os.fstat(fh.fileno())[stat.ST_SIZE]
            filename = fh.name.split('/')[-1]
            contenttype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
            body += '--%s\r\n' % boundary
            body += 'Content-Disposition: form-data; name="%s"; filename="%s"\r\n' % (key, filename)
            body += 'Content-Type: %s\r\n' % contenttype
            body += 'Content-Length: %s\r\n' % file_size
            fh.seek(0)
            body += '\r\n' + fh.read() + '\r\n'
            fh.close()
        body += '--%s--\r\n\r\n' % boundary
        return boundary, body

    def put(self, url, body=None, contentType='application/xml', accept='application/xml'):
        return self._httpCall(url, 'PUT', body, contentType, accept)

    def delete(self, url, body=None, contentType='application/x-www-form-urlencoded', accept='application/xml'):
        return self._httpCall(url, 'DELETE', body, contentType, accept)

    def head(self, url):
        return self._httpCall(url, 'HEAD')

    def addCredentials(self, username, password):
        self.crendentials[username] = password

    def addCertificate(self, key, cert):
        self.certificates[key] = cert

    def setHandleResponse(self, handle):
        self.handleResponse = handle

    def _addCredentials(self, http):
        for u, p in self.crendentials.items():
            http.add_credentials(u, p)

    def _addCredentialsToHeader(self, headers):
        if self.crendentials:
            u, p = self.crendentials.items()[0]
            headers['authorization'] = 'Basic ' + base64.b64encode('%s:%s' % (u, p))
            return headers

    def _addCertificate(self, http):
        for u, p in self.certificates.items():
            http.add_certificate(u, p, '')

    def _printDetail(self, message):
        Util.printDetail(message, self.verboseLevel, Util.VERBOSE_LEVEL_DETAILED)

    # This is a bandaid for problems associated with dropped SSL handshakes.  The
    # root cause of these problems needs to be found and fixed.
    def _retryHttpRequestOnSSLError(self, httpObject, url, method, body, headers):
        maxRetries = 3
        retries = 0
        lastException = None
        while retries < maxRetries:
            try:
                if len(headers):
                    return httpObject.request(url, method, body, headers=headers)
                else:
                    return httpObject.request(url, method, body)
            except ssl.SSLError as e:
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                self._printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                lastException = e
                retries += 1
            except httplib2.ssl_SSLError as e:
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                self._printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                lastException = e
                retries += 1

        raise lastException

    def _httpCall(self, url, method, body=None, contentType='application/xml', accept='application/xml', retry=True):

        def _convertContent(content):

            size = len(content)
            if size > 2048:
                return '<content too large; %d bytes>' % size

            try:
                return unicode(content, 'utf-8')
            except:
                return '<non-text content>'

        def _getErrorMessageFromJsonContent(content):
            try:
                return json.loads(content)['message']
            except:
                return ''

        def _handle3xx():
            if resp.status == 302:
                # Redirected
                resp, content = self._httpCall(resp['location'], method, body, accept)
            else:
                raise Exception('Should have been handled by httplib2!! ' + str(resp.status) + ": " + resp.reason)
            return resp, content

        def _handle4xx():
            error_message = _getErrorMessageFromJsonContent(content)
            raise ClientException('Failed calling method %s on url %s, with reason: %s. Error: %s' %
                                  (method, url, str(resp.status) + ": " + resp.reason, error_message),
                                  content=content,
                                  status=str(resp.status))

        def _handle5xx():
            if retry:
                return self._httpCall(url, method, body, contentType, accept, False)
            raise ServerException('Failed calling method %s on url %s, with reason: %s' %
                                         (method, url, str(resp.status) + ": " + resp.reason),
                                         status=str(resp.status))

        def _handleResponse(resp, content):
            self._printDetail('Received response: %s' % resp + \
                             '\nwith content:\n %s' % \
                             _convertContent(content))

            if str(resp.status).startswith('2'):
                return resp, content

            if str(resp.status).startswith('3'):
                resp, content = _handle3xx()

            if str(resp.status).startswith('4'):
                resp, content = _handle4xx()

            if str(resp.status).startswith('5'):
                resp, content = _handle5xx()

        if self.useHttpCache:
            h = httplib2.Http(".cache")
        else:
            h = httplib2.Http()
        h.force_exception_to_status_code = False
        h.disable_ssl_certificate_validation=True
        self._printDetail('Contacting the server with %s, at: %s' % (method, url))
        headers = {}
        if contentType:
            headers['Content-Type'] = contentType
        if accept:
            headers['Accept'] = accept
        # See https://github.com/StratusLab/client/issues/8
        if method == 'POST':
            self._addCredentialsToHeader(headers)
        self._addCredentials(h)
        self._addCertificate(h)
        try:
            resp, content = self._retryHttpRequestOnSSLError(h, url, method, body, headers)
        except httplib.BadStatusLine:
            raise NetworkException('BadStatusLine when contacting ' + url)
        except AttributeError:
            raise NetworkException('Cannot contact ' + url)

        if self.handleResponse:
            try:
                _handleResponse(resp, content)
            except ClientException, ex:
                ex.mediaType = headers['Accept']
                raise

        return resp, content




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import os
import ConfigParser
from stratuslab import Defaults

from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Exceptions import ValidationException, InputException
from stratuslab.marketplace.ManifestDownloader import ManifestDownloader

class Filter(object):
    
    def __init__(self, messages, config):
        self.filterItems = []
        self.messages = messages
        self.successMessage = None
        self.failedMessage = None
        self.config = config

    def _loadConfig(self, sectionName, configParameterName):
        try:
            rawFilterItems = self.config.get(sectionName, configParameterName)
            self.filterItems = filter(None,[item.strip() for item in rawFilterItems.split('\n')])
        except ConfigParser.NoOptionError:
            pass

    def filter(self, metadatas):

        for metadata in metadatas:
            candidate = self._getCandidate(metadata)
            if self._filterSingle(candidate) is False:
                metadatas.remove(metadata)

        return metadatas

    def _getCandidate(self, manifestInfo):
        pass
    
    def _filterSingle(self, candidate):
        pass
    

class BlackListBaseFilter(Filter):

    def _filterSingle(self, candidate):
        if candidate not in self.filterItems:
            self.messages.append(self.successMessage % candidate)
            return True
        else:
            self.messages.append(self.failedMessage % candidate)
            return False

class WhiteListBaseFilter(Filter):

    def _filterSingle(self, candidate):
        if candidate in self.filterItems:
            self.messages.append(self.successMessage % candidate)
            return True
        else:
            self.messages.append(self.failedMessage % candidate)
            return False

class WhiteListEndorsersFilter(WhiteListBaseFilter):
    
    def __init__(self, messages, config):
        super(WhiteListEndorsersFilter, self).__init__(messages, config)
        self._loadConfig('endorsers', 'whitelistendorsers')
        self.successMessage = 'email endorser %s is whitelisted, keeping entry'
        self.failedMessage = 'email endorser %s is not whitelisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.endorser


class BlackListEndorsersFilter(BlackListBaseFilter):
    
    def __init__(self, messages, config):
        super(BlackListEndorsersFilter, self).__init__(messages, config)
        self._loadConfig('endorsers', 'blacklistendorsers')
        self.successMessage = 'email endorser %s is not blacklisted, keeping entry'
        self.failedMessage = 'email endorser %s is blacklisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.endorser


class WhiteListChecksumsFilter(WhiteListBaseFilter):
    
    def __init__(self, messages, config):
        super(WhiteListChecksumsFilter, self).__init__(messages, config)
        self._loadConfig('checksums', 'whitelistchecksums')
        self.successMessage = 'SHA-1 checksum image %s is whitelisted, keeping entry'
        self.failedMessage = 'SHA-1 checksum image %s is not whitelisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.sha1


class BlackListChecksumsFilter(BlackListBaseFilter):
    
    def __init__(self, messages, config):
        super(BlackListChecksumsFilter, self).__init__(messages, config)
        self._loadConfig('checksums', 'blacklistchecksums')
        self.successMessage = 'SHA-1 checksum image %s is not blacklisted, keeping entry'
        self.failedMessage = 'SHA-1 checksum image %s is blacklisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.sha1


class WhiteListImagesFilter(WhiteListBaseFilter):
    
    def __init__(self, messages, config):
        super(WhiteListImagesFilter, self).__init__(messages, config)
        self._loadConfig('images', 'whitelistimages')
        self.successMessage = 'image identifier %s is whitelisted, keeping entry'
        self.failedMessage = 'image identifier %s is not whitelisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.identifier


class BlackListImagesFilter(BlackListBaseFilter):
    
    def __init__(self, messages, config):
        super(BlackListImagesFilter, self).__init__(messages, config)
        self._loadConfig('images', 'blacklistimages')
        self.successMessage = 'image identifier %s is not blacklisted, keeping entry'
        self.failedMessage = 'image identifier %s is blacklisted, rejecting entry'

    def _getCandidate(self, manifestInfo):
        return manifestInfo.identifier


class Policy(object):

    POLICY_CFG = os.path.join(Defaults.ETC_DIR, 'policy.cfg')

    def __init__(self, configHolder = ConfigHolder()):
        self.messages = []
        self.policyConfigFile = None
        self.configHolder = configHolder
        configHolder.assign(self)
        self.filters = self._loadFilters()

    def _loadFilters(self):
        filters = []

        configFile = self.policyConfigFile or Policy.POLICY_CFG
        if not os.path.exists(configFile):
            raise InputException("Can't find policy configuration file: %s" % configFile)
        config = ConfigParser.ConfigParser()
        config.read(configFile)

        whiteListEndorsersFilter = WhiteListEndorsersFilter(self.messages, config)
        if whiteListEndorsersFilter.filterItems:
            filters.append(whiteListEndorsersFilter)

        blackListEndorsersFilter = BlackListEndorsersFilter(self.messages, config)
        if blackListEndorsersFilter.filterItems:
            filters.append(blackListEndorsersFilter)

        whiteListChecksumsFilter = WhiteListChecksumsFilter(self.messages, config)
        if whiteListChecksumsFilter.filterItems:
            filters.append(whiteListChecksumsFilter)

        blackListChecksumsFilter = BlackListChecksumsFilter(self.messages, config)
        if blackListChecksumsFilter.filterItems:
            filters.append(blackListChecksumsFilter)

        whiteListImagesFilter = WhiteListImagesFilter(self.messages, config)
        if whiteListImagesFilter.filterItems:
            filters.append(whiteListImagesFilter)

        blackListImagesFilter = BlackListImagesFilter(self.messages, config)
        if blackListImagesFilter.filterItems:
            filters.append(blackListImagesFilter)

        return filters

    def check(self, identifierUri):

        manifests = self._downloadManifests(identifierUri)

        remainingMetadataEntries = manifests
        
        for filter in self.filters:
            remainingMetadataEntries = filter.filter(remainingMetadataEntries)

        if len(remainingMetadataEntries) == 0:
            sys.stderr.write(self._errorMessage())
            raise ValidationException('Policy check Failed')
        
        return remainingMetadataEntries

    def _downloadManifests(self, identifier):
        return ManifestDownloader(self.configHolder).getManifestList(identifier)

    def _errorMessage(self):
        self.messages.append("Image isn't valid according to site policy\n")
        return '\n'.join(self.messages)






#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.HttpClient import HttpClient
from stratuslab.Exceptions import InputException, ClientException,\
    ExecutionException
from stratuslab.ManifestInfo import ManifestInfo
from stratuslab.marketplace.Util import Util as MarketplaceUtil
from stratuslab.Util import importETree

etree = importETree()

class Uploader(object):

    def __init__(self, configHolder = ConfigHolder()):
        self.confHolder = configHolder
        configHolder.assign(self)

    @staticmethod
    def buildUploadParser(parser):
        MarketplaceUtil.addEndpointOption(parser)


    @staticmethod
    def checkUploadOptions(options, parser):
        MarketplaceUtil.checkEndpointOption(options)


    def upload(self, manifestFilename):
        client = HttpClient()
        if not os.path.exists(manifestFilename):
            raise InputException('Can\'t find metadata file: %s' % manifestFilename)
        
        manifest = open(manifestFilename).read()

        info = ManifestInfo(self.confHolder)
        info.parseManifest(manifest)
        
        url = MarketplaceUtil.metadataEndpointUrl(self.marketplaceEndpoint)
        try:
            client.post(url, manifest)
        except ClientException, ex:
            error = ''
            try:
                error = etree.fromstring(ex.content).text
            except: pass
            raise ExecutionException("Failed to upload: %s: %s" % (ex.reason, error))
        except AttributeError, ex:
            raise ExecutionException("Failed to upload (post) to URL: %s" % url)

        finalUrl = MarketplaceUtil.metadataCompleteUrl(self.marketplaceEndpoint, 
                                                       info.identifier, 
                                                       info.email, 
                                                       info.created)
        return finalUrl



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os.path

import stratuslab.Util as Util
from stratuslab.ManifestInfo import ManifestInfo
from stratuslab.Exceptions import ExecutionException
from stratuslab.Signator import Signator
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Compressor import Compressor
from stratuslab.marketplace.Util import Util as MarketplaceUtil
from stratuslab.marketplace.Uploader import Uploader as MarketplaceUploader
from stratuslab.PersistentDisk import PersistentDisk
from stratuslab.commandbase.StorageCommand import PDiskEndpoint, PDiskVolume


class Uploader(object):
    @staticmethod
    def buildUploadParser(parser):
        parser.usage = '''usage: %prog [options] <image-file>'''

        parser.add_option('-C', '--compress', dest='compressionFormat',
                          help='Compression format.  One of: %s, none' % ', '.join(Compressor.compressionFormats),
                          default='gz', metavar='FORMAT')

        parser.add_option('--image-only', dest='imageOnly',
                          help='Do not upload metadata file to Marketplace',
                          action='store_true', default=False)

        MarketplaceUtil.addEndpointOption(parser)

        PDiskEndpoint.addOptions(parser)
        PDiskVolume.addOptions(parser)

    @staticmethod
    def checkUploadOptions(options, parser):

        if options.marketplaceEndpoint:
            options.withMarketPlace = True

        if not options.imageOnly:
            MarketplaceUtil.checkEndpointOption(options)

        allowedFormats = list(Compressor.compressionFormats)
        allowedFormats.append('none')

        if options.compressionFormat not in allowedFormats:
            parser.error('Unknown compression format')

        # Do NOT check the validity of the pdisk options, so that the
        # values can be taken from non-pdisk options when necessary.
        # PDiskEndpoint.checkOptions(options)

        # This essentially just checks that the volume tag is acceptable.
        PDiskVolume.checkOptions(options)

    def __init__(self, imageFile, configHolder=ConfigHolder()):
        self.imageMetadata = {}

        self.configHolder = configHolder
        configHolder.assign(self)

        self.imageFile = imageFile
        self.manifestFile = self.imageFile.replace('.img', '.xml')

        self.imageUrl = ''

        self.pdisk = PersistentDisk(self.configHolder)

    def start(self):
        Util.printAction('Starting image upload')

        Util.printStep('Compressing image')
        self._compressImage()

        Util.printStep('Uploading image')
        self._uploadImage()

        if not self.imageOnly:
            Util.printAction('Starting manifest upload')

            Util.printStep('Parsing manifest')
            self._parseManifest()

            Util.printStep('Updating manifest')
            self._updateManifest()

            Util.printStep('Signing manifest')
            self._signManifest()

            Util.printStep('Uploading manifest')
            self._uploadMarketplaceManifest()

    def _uploadImage(self):
        self.imageUrl = self.pdisk.uploadVolume(self.imageFile)
        Util.printInfo('Image uploaded: %s' % self.imageUrl)

        self._updateImageMetadataInPDisk()

    def _updateImageMetadataInPDisk(self):
        if self.imageMetadata:
            uuid = self.imageUrl.rsplit('/', 1)[-1]
            self.pdisk.updateVolumeAsUser(self.imageMetadata, uuid)
            Util.printInfo('Image metadata updated: %s' % self.imageMetadata)

    def _updateManifest(self):
        self._addLocationToManifest()
        self._addCompressionFormatToManifest()

    def _signManifest(self):
        configHolder = ConfigHolder(self.__dict__)
        signator = Signator(self.manifestFile, configHolder)
        rc = signator.sign()
        if rc:
            raise ExecutionException('Failed to sign manifest.')
        self.manifestFile = signator.outputManifestFile

    def _execute(self, command):
        if self.verboseLevel <= Util.VERBOSE_LEVEL_NORMAL:
            devNull = open(os.path.devnull, 'w')
            ret = Util.execute(command, stdout=devNull, stderr=devNull)
            devNull.close()
        else:
            ret = Util.execute(command)
        return ret

    def _parseManifest(self):
        manifestInfo = ManifestInfo()
        manifestInfo.parseManifestFromFile(self.manifestFile)

    def _buildManifestName(self, repoFilename):
        return repoFilename.split('.img')[0] + '.xml'

    def _compressFile(self, filename, fmt):

        if fmt.lower() == 'none':
            return filename

        if Compressor.getCompressionFormat(filename) != '':
            Util.printWarning('skipping compression; file appears to already be compressed')
            return filename

        compressionCmd = Compressor._getCompressionCommandByFormat(fmt)

        compressedFilename = '%s.%s' % (filename, fmt)
        if os.path.isfile(compressedFilename):
            Util.printWarning('Compressed file %s already exists, skipping' % compressedFilename)
            return compressedFilename

        if not os.path.exists(filename):
            Util.printError('Missing file: ' + filename, exit=True)

        ret = self._execute([compressionCmd, filename])
        if ret != 0:
            Util.printError('Error compressing file: %s' % compressedFilename, exit=True)

        return compressedFilename

    def _compressImage(self):
        self.imageFile = self._compressFile(self.imageFile, self.compressionFormat)

    def _addCompressionFormatToManifest(self):
        if self.compressionFormat.lower() != 'none':
            ManifestInfo.addElementToManifestFile(self.manifestFile,
                                                  'compression', self.compressionFormat)
        else:
            Util.printWarning("'none' specified for compression; manifest compression element NOT updated")

    def _addLocationToManifest(self):
        ManifestInfo.addElementToManifestFile(self.manifestFile,
                                              'location', self.imageUrl)

    def _uploadMarketplaceManifest(self):
        uploader = MarketplaceUploader(self.configHolder)
        url = uploader.upload(self.manifestFile)
        Util.printInfo('Manifest uploaded: %s' % url)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2012, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import stomp

from stratuslab.messaging.MsgBase import MsgBase

STOMP_PORT = '61613'

class StompClient(MsgBase):
    def __init__(self, configHolder):
        super(StompClient, self).__init__(configHolder)
        
        port = STOMP_PORT
        try:
            host, port = self.msg_endpoint.split(':')
        except ValueError:
            host = self.msg_endpoint

        self.connection = stomp.Connection(host_and_ports=[(host, int(port))])

    def connect(self):
        self.connection.start()
        self.connection.connect()

    def send(self, message):
        if not self.connection.connected:
            self.connect()

        self.connection.subscribe(destination=self.msg_queue, ack='auto')
        self.connection.send(message, destination=self.msg_queue)

    def disconnect(self):
        self.connection.disconnect()



from distutils.core import setup

import os
import os.path

try:
    rootdir = os.path.join('lib','stratuslab','python')
    for file in os.listdir(rootdir):
        src = os.path.join(rootdir, file)
        os.symlink(src, file)
except:
    # assume links have already been made
    pass

setup(
    name='stratuslab-client',
    version='13.05.0.RC1',
    author='StratusLab',
    author_email='contact@stratuslab.eu',
    url='http://stratuslab.eu/',
    license='Apache Software License 2.0',
    description='StratusLab command line client for users',
    long_description=open('README.txt').read(),

    scripts=[
        'bin/stratus-attach-volume',
        'bin/stratus-build-metadata',
        'bin/stratus-connect-instance',
        'bin/stratus-copy-config',
        'bin/stratus-create-image',
        'bin/stratus-create-volume',
        'bin/stratus-delete-volume',
        'bin/stratus-deprecate-metadata',
        'bin/stratus-describe-instance',
        'bin/stratus-describe-volumes',
        'bin/stratus-detach-volume',
        'bin/stratus-hash-password',
        'bin/stratus-kill-instance',
        'bin/stratus-prepare-context',
        'bin/stratus-run-cluster',
        'bin/stratus-run-instance',
        'bin/stratus-shutdown-instance',
        'bin/stratus-sign-metadata',
        'bin/stratus-update-volume',
        'bin/stratus-upload-image',
        'bin/stratus-upload-metadata',
        'bin/stratus-validate-metadata',
        ],

     packages=[
        'stratuslab',
        'stratuslab.api',
        'stratuslab.cloud',
        'stratuslab.cloudinit',
        'stratuslab.commandbase',
        'stratuslab.image',
        #'stratuslab.installator',
        'stratuslab.marketplace',
        'stratuslab.messaging',
        'stratuslab.pat',
        'stratuslab.system',
        #'stratuslab.tm',
        ],

    data_files=[
        ('java', ['java/metadata-13.05.0.RC1-jar-with-dependencies.jar']),
        ('share/vm', ['share/vm/schema.one']),
        ('share/template', ['share/template/manifest.xml.tpl']),
        ('conf', ['conf/stratuslab-user.cfg.ref']),
        ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 2.6',
        'Topic :: System :: Distributed Computing',
        ],

    install_requires=[
        "dirq >= 1.2.2",
        "stomp.py >= 3.1.3",
        "pika >= 0.9.9",
        "httplib2 >= 0.7.7",
        "couchbase == 0.8.2"
    ],
)



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2012, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.CommandBase import CommandBaseSysadmin
from stratuslab.messaging.Defaults import MSG_TYPES
from stratuslab.messaging.Defaults import MSG_SMTP_HOST
from stratuslab.messaging.MessagePublishers import ImageIdPublisher, SingleMessagePublisher
from stratuslab.image.Image import Image
from stratuslab.Exceptions import ValidationException


class MessagingCommand(CommandBaseSysadmin):
    def __init__(self):
        self.msg_message = ''
        super(MessagingCommand, self).__init__()

    def parse(self):
        self.parser.usage = '%prog [options] message'

        self.parser.description = '''
Send the given message to the given messaging server.
'''

        self.parser.add_option('--msg-type', dest='msg_type',
                               help='Type of messaging: %s. Mandatory.' % ', '.join(MSG_TYPES),
                               metavar='NAME', default="")
        self.parser.add_option('--msg-endpoint', dest='msg_endpoint',
                               help='Messaging service endpoint. Mandatory.',
                               metavar='ENDPOINT', default="")
        self.parser.add_option('--msg-queue', dest='msg_queue',
                               help='Message queue name. Mandatory.', metavar='NAME',
                               default="")
        self.parser.add_option('--msg-smtp-host', dest='smtp_host',
                               help="SMTP relay hostname to be used with 'email' messaging "
                                    "type. Default: %s" % MSG_SMTP_HOST,
                               metavar='HOSTNAME', default=MSG_SMTP_HOST)
        self.parser.add_option('--imageid', dest='imageid',
                               help='Image ID. Assumes message is JSON representation '
                                    'of a dictionary on which the ID will be set. JSON can be '
                                    'base64 encoded.', metavar='ID',
                               default="")

        self.options, self.args = self.parser.parse_args()

    def checkOptions(self):
        super(MessagingCommand, self).checkOptions()

        if not self.options.msg_type:
            self.printMandatoryOptionError('--msg-type')
        if not self.options.msg_endpoint:
            self.printMandatoryOptionError('--msg-endpoint')
        if not self.options.msg_queue:
            self.printMandatoryOptionError('--msg-queue')

        self._checkMessageAndImageid()

    def _checkMessageAndImageid(self):
        if len(self.args) != 0:
            self.msg_message = self.args[0]
        else:
            self.printError('Message should be set as first argument')

        if self.options.imageid:
            if not Image.isImageId(self.options.imageid):
                raise ValidationException('Marketplace image ID is expected.')

    def sendMessage(self):
        config = ConfigHolder.configFileToDict(self.options.configFile)
        configHolder = ConfigHolder(self.options.__dict__, config)

        if self.options.imageid:
            ImageIdPublisher(self.msg_message,
                             self.options.imageid, configHolder).publish()
        else:
            SingleMessagePublisher(self.msg_message, configHolder).publish()



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import time
import ssl
from time import gmtime, strftime
from stratuslab.Exceptions import OneException
from stratuslab import Util

etree = Util.importETree()


class OneConnector(object):

    ACL_USERS = {
        "UID": 0x100000000,
        "GID": 0x200000000,
        "ALL": 0x400000000
    }
    ACL_RESOURCES = {
        "VM": 0x1000000000,
        "HOST": 0x2000000000,
        "NET": 0x4000000000,
        "IMAGE": 0x8000000000,
        "USER": 0x10000000000,
        "TEMPLATE": 0x20000000000,
        "GROUP": 0x40000000000
    }
    ACL_RIGHTS = {
        "USE": 0x1,  # Auth. to use an object
        "MANAGE": 0x2,  # Auth. to perform management actions
        "ADMIN": 0x4,  # Auth. to perform administrative actions
        "CREATE": 0x8   # Auth. to create an object
    }

    def __init__(self, credentials):

        self._sessionString = None
        self._rpc = None
        self._credentials = credentials

    def setEndpointFromParts(self, server, port=2634, path='xmlrpc', protocol='https'):
        self._setEndpointFromParts(server, port, path, protocol)
        self.setEndpoint(self.server)

    def _setEndpointFromParts(self, server, port=2634, path='xmlrpc', protocol='https'):
        self.server = '%s://%s:%s/%s' % (protocol, server, port, path)

    def setEndpoint(self, address):
        if not address:
            raise ValueError('Missing endpoint')
        if not address.startswith('http'):
            self._setEndpointFromParts(address)
        else:
            self.server = address
        self._createRpcConnection()
        return self.server

    def _createRpcConnection(self):
        self._rpc = self._credentials.createRpcConnection()
        self._sessionString = self._credentials.createSessionString()

    def vmStart(self, vmTpl):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                isSuccess, detail, _ = self._rpc.one.vm.allocate(self._sessionString, vmTpl)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        self._raiseIfError(isSuccess, detail)

        vmId = detail

        return vmId

    def vmStop(self, vmId):
        self._vmAction(vmId, 'shutdown')

    def vmKill(self, vmId):
        self._vmAction(vmId, 'finalize')

    def _vmAction(self, vmId, action):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                res = self._rpc.one.vm.action(self._sessionString, action, vmId)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        isSuccess = res[0]
        detail = ''
        if len(res) > 1:
            detail = res[1]

        self._raiseIfError(isSuccess, detail)

        return

    def _raiseIfError(self, isSuccess, reason):
        if not isSuccess:
            raise OneException(reason)

    def listVms(self, showVmsFromAllUsers=False):
        fromAllUsers = -2
        currentUserOnly = -3

        if showVmsFromAllUsers:
            visibilitySwitch = fromAllUsers
        else:
            visibilitySwitch = currentUserOnly

        # Hack to retry on SSL errors.
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.vmpool.info(self._sessionString, visibilitySwitch, -1, -1, -1)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        vmlist = etree.fromstring(info)
        for xml in vmlist.findall('VM'):
            self._addStateSummary(xml)

        return etree.tostring(vmlist)

    def _getVmInfoAsXml(self, vmId):
        info = self._vmInfo(vmId)

        xml = etree.fromstring(info)

        self._addStateSummary(xml)

        return xml

    def _addStateSummary(self, xml):
        vmState = self._getOneVmStateFromXml(xml)

        labelElement = etree.Element('STATE_SUMMARY')
        labelElement.text = str(vmState)
        xml.append(labelElement)

    def _vmInfo(self, vmId):

        # Hack to retry on SSL errors.
        maxRetries = 3
        retries = 0
        while True:
            try:
                isSuccess, info, _ = self._rpc.one.vm.info(self._sessionString, vmId)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        self._raiseIfError(isSuccess, info)
        return info

    def getVmInfo(self, vmId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        self._addStateSummary(xml)
        return etree.tostring(xml)

    def getVmNode(self, vmId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        return xml.find('HISTORY_RECORDS/HISTORY/HOSTNAME').text

    def getVmOwner(self, vmId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        return xml.find('UNAME').text

    def getVmDiskSource(self, vmId, diskId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        sources = [x.find('SOURCE').text for x in xml.findall('TEMPLATE/DISK')
                   if x.find('DISK_ID').text == str(diskId)]
        return sources[0]

    def getCreateImageInfo(self, vmId):
        info = self._vmInfo(vmId)
        dom = etree.fromstring(info)
        createImage_dom = dom.find('TEMPLATE/CREATE_IMAGE')
        infos = {}
        for elem in createImage_dom:
            value = elem.text or ''
            infos[elem.tag] = value.strip().strip('"')
        return infos

    def _findXmlText(self, xml, query):
        return xml.find(query).text.strip('"')

    def getVmSource(self, vmId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        return xml.find('TEMPLATE/DISK/DISK_ID=0/../SOURCE').text

    def isVmRunning(self, vmId):
        return str(self._getVmStateSummary(vmId)) == 'Running'

    def _getOneVmStateFromXml(self, xml):
        stateElement = self._getStateElement(xml)
        state = int(stateElement.text)

        stateElement = self._getLcmStateElement(xml)
        lcmState = int(stateElement.text)

        return OneVmState(state, lcmState)

    def _getVmLcmStateLabel(self, vmId):
        status = self._getVmLcmState(vmId)
        return self.lcmStatusLabel[status]

    def _getVmLcmState(self, vmId):
        xml = self._getVmStateAsXml(vmId)
        status = self._getLcmStateElement(xml)
        return int(status.text)

    def _getStateElement(self, xml):
        return xml.find('STATE')

    def _getLcmStateElement(self, xml):
        return xml.find('LCM_STATE')

    def getVmIp(self, vmId):
        xml = self._getVmInfoAsXml(vmId)

        nic = xml.find('TEMPLATE/NIC')
        networkName = nic.find('NETWORK').text
        ip = nic.find('IP').text

        return networkName, ip

    def getVmSshPort(self, *args, **kwargs):
        return 22

    def getVmState(self, vmId):
        return str(self._getVmStateSummary(vmId))

    def waitUntilVmRunningOrTimeout(self, vmId, timeout, ticks=True, failOn=()):
        start = time.time()
        state = ''

        noWaitStates = ('Running', 'Failed')

        while str(state) not in noWaitStates:

            state = self._getVmStateSummary(vmId)

            if str(state) in failOn:
                return False

            if ticks:
                sys.stdout.flush()
                sys.stdout.write('.')
            time.sleep(1)

            if time.time() - start > timeout:
                if ticks:
                    sys.stdout.flush()
                    sys.stdout.write('\n')
                return False

        if ticks:
            sys.stdout.flush()
            sys.stdout.write('\n')
        return True

    def _getVmStateSummary(self, vmId):
        info = self._vmInfo(vmId)
        xml = etree.fromstring(info)
        return self._getOneVmStateFromXml(xml)

    # -------------------------------------------
    #    Virtual network management
    # -------------------------------------------

    def networkCreate(self, vnetTpl):

        # Hack to retry on SSL errors.
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, vnetId, _ = self._rpc.one.vn.allocate(self._sessionString, vnetTpl)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            error = vnetId
            raise OneException('Error creating ONE network:\n%s' % error)

        return vnetId

    def getNetworkPoolInfo(self, filter=-2):

        # Hack to retry on SSL errors.
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.vnpool.info(self._sessionString, filter)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info

    def getNetworkInfo(self, vnetId):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.vn.info(self._sessionString, vnetId)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info

    # -------------------------------------------
    #    Host management
    # -------------------------------------------

    def hostCreate(self, hostname, im, vmm, tm, vnm='dummy', inDomain=True):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, id, _ = self._rpc.one.host.allocate(self._sessionString, hostname, im, vmm, vnm, tm, inDomain)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(id)

        return id

    def hostRemove(self, id):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret = self._rpc.one.host.delete(self._sessionString, id)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(id)

        return id

    def getHostInfo(self, id):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.host.info(self._sessionString, id)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info

    def listHosts(self):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.hostpool.info(self._sessionString)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info

    # -------------------------------------------
    #    ACLs management
    # -------------------------------------------

    def addNetworkAcl(self, users, net_id_int, rights):
        """
        users - hex
        net_id_int - integer, network ID
        rights - hex
        """
        # "magic" number
        _magic = self.ACL_USERS['UID']
        net_resource = hex(self.ACL_RESOURCES['NET'] + _magic + net_id_int)

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.acl.addrule(self._sessionString,
                                                         users,
                                                         net_resource,
                                                         rights)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info

    def addUserAcl(self, users, resources, rights):

        # Hack to retry on SSL errors
        maxRetries = 3
        retries = 0
        while True:
            try:
                ret, info, _ = self._rpc.one.acl.addrule(self._sessionString,
                                                         users,
                                                         resources,
                                                         rights)
                break
            except ssl.SSLError as e:
                retries += 1
                t = strftime("%Y-%m-%d %H:%M:%S", gmtime())
                Util.printDetail('SSL ERROR ENCOUNTERED (%s): %s' % (t, str(e)))
                if retries >= maxRetries:
                    raise e

        if not ret:
            raise OneException(info)

        return info


class OneVmState(object):

    def __init__(self, state, lcmState=None):
        self. state = int(state)

        self.lcmState = lcmState

        self.invalidState = 'Invalid state'

        self.stateDefinition = ['init',
                                'pending',
                                'hold',
                                'active',
                                'stopped',
                                'suspended',
                                'done',
                                'failed']

        self.lcmStateDefintion = ['lcm_init',
                                  'prolog',
                                  'boot',
                                  'running',
                                  'migrate',
                                  'save_stop',
                                  'save_suspend',
                                  'save_migrate',
                                  'prolog_migrate',
                                  'prolog_resume',
                                  'epilog_stop',
                                  'epilog',
                                  'shutdown',
                                  'cancel',
                                  'failure',
                                  'delete',
                                  'unknown']

    def __str__(self):
        if self._useLcmState():
            s = self._lcmStateToString()
        else:
            s = self._stateToString()
        return s.title()

    def _useLcmState(self):
        stateForLcmStateLookup = 3
        return self.state == stateForLcmStateLookup

    def _stateToString(self):
        if (self.state < 0) and (self.state >= len(self.stateDefinition)):
            return self.invalidState
        return self.stateDefinition[self.state]

    def _lcmStateToString(self):
        lcm = self._lcmStateAsInt()
        if (lcm is not None) and (lcm >= 0) and (lcm < len(self.lcmStateDefintion)):
            return self.lcmStateDefintion[lcm]
        else:
            Util.printError('Invalid state: %s' % lcm, exit=False)
            return self.invalidState

    def _lcmStateAsInt(self):
        lcm = None
        if self.lcmState:
            lcm = int(self.lcmState)
        return lcm


class OneHostState(object):

    def __init__(self, state):
        self. state = state
        self.invalidState = 'Invalid state'
        self.stateDefinition = {'0': 'INIT',
                                '1': 'MONITORING',
                                '2': 'MONITORED',
                                '3': 'ERROR',
                                '4': 'DISABLED'}

    def __str__(self):
        return self.stateDefinition.get(self.state, self.invalidState).title()



from stratuslab import Exceptions

INSTALLERS = ('yum', 'apt')

yumOsList = ('centos', 'fedora', 'sl', 'scientificlinux')
aptOsList = ('ubuntu')

def getInstallerBasedOnOs(os):
    if os in yumOsList:
        return 'yum'
    elif os in aptOsList:
        return 'apt'
    else:
        raise Exceptions.ExecutionException('Unsupported OS: %s' % os)

packageAndVersionSeparators = {'yum' : '-',
                               'apt' : '='}

def getPackageAndVersionSeparatorBasedOnOs(os):
    return packageAndVersionSeparators[getInstallerBasedOnOs(os)]



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

from stratuslab.Runner import Runner
from stratuslab.AuthnCommand import AuthnCommand
import stratuslab.Util as Util

from marketplace.Util import Util as MarketplaceUtil

class Runnable(AuthnCommand):
    '''Base class for command which need to start a machine.'''
    parser_usage = '''%prog [options] image'''
    parser_description = 'image - Marketplace image ID or PDISK volume UUID'

    def __init__(self):
        self.options = None
        self.args = None
        self.image = None

        super(Runnable, self).__init__()

    def parse(self):
        defaultOptions = Runner.defaultRunOptions()

        self.parser.usage = self.parser_usage

        self.parser.description = self.parser_description

        self.parser.add_option('-k', '--key', dest='userPublicKeyFile',
                help='SSH public key(s) (.pub) to log on the machine. Default %s. In case of multiple keys, concatenate them to the file.' % defaultOptions['userPublicKeyFile'], 
                metavar='FILE',
                default=defaultOptions['userPublicKeyFile'])

        self.parser.add_option('-t', '--type', dest='instanceType',
                help='instance type to start (see --list-types for default)', metavar='TYPE',
                default=Runner.DEFAULT_INSTANCE_TYPE)

        self.parser.add_option('--list-type', dest='listType',
                help='list available instance types (deprecated)',
                action='store_true')

        self.parser.add_option('-l', '--list-types', dest='listType',
                help='list available instance types',
                default=False, action='store_true')

        self.parser.add_option('--cpu', dest='vmCpu',
                help='number of CPU cores',
                default=None)

        self.parser.add_option('--ram', dest='vmRam',
                help='RAM in megabytes',
                default=None)

        self.parser.add_option('--swap', dest='vmSwap',
                help='swap space in megabytes',
                default=None)

        self.parser.add_option('--context-file', dest='extraContextFile', metavar='FILE',
                help='extra context file with one key=value per line',
                default=defaultOptions['extraContextFile'])
        self.parser.add_option('--context', dest='extraContextData', metavar='CONTEXT',
                help='extra context string (separate by %s)' % Util.cliLineSplitChar,
                default=defaultOptions['extraContextData'])

        self.parser.add_option('--cloud-init', dest='cloudInit', metavar='PAIRS',
                help='mimetype,file pairs (separate by %s)' % Util.cliLineSplitChar,
                default=defaultOptions['cloudInit'])

        self.parser.add_option('--vnc-port', dest='vncPort', metavar='PORT', type='int',
                help='VNC port number. Note for KVM it\'s the real one, not the '
                     'VNC port. So for VNC port 0 you should specify 5900, for '
                     'port 1 is 5901 and so on. ',
                default=defaultOptions['vncPort'])
        self.parser.add_option('--vnc-listen', dest='vncListen', metavar='ADDRESS',
                help='IP to listen on',
                default=defaultOptions['vncListen'])

        self.parser.add_option('--vm-requirements', dest='vmRequirements', metavar='REQUIREMENTS',
                help='Advanced requirements for VM placement. '
                     'Available values depend on the cloud configuration. '
                     'e.g. --vm-requirements \'CLUSTER = \\"test\\" & MAGICNUMBER = 42\'',
                default=None)

        self.parser.add_option('--vm-template-file', dest='vmTemplateFile', metavar='FILE',
                help='VM template file. Default %s' % defaultOptions['vmTemplateFile'],
                default=defaultOptions['vmTemplateFile'])

        self.parser.add_option('--vm-cpu-amount', dest='vmCpuAmount', metavar='CPU', type='float',
                help='Percentage of CPU divided by 100 required for the Virtual Machine. '
                     'Half a processor is written 0.5. No default. If not provided, CPU value from '
                     'predefined instance types is used.',
                default=defaultOptions['vmCpuAmount'])

        self.parser.add_option('--vm-disks-bus', dest='vmDisksBus', metavar='BUSTYPE',
                help='VM disks bus type defined for all disks. Overrides "disks-bus" '
                'element value defined in image manifest. '
                'Available types: %s. ' % ', '.join(Runner.DISKS_BUS_AVAILABLE) + 
                'If not provided, by default the value is taken from disks-bus '
                'element of image manifest. If the latter is not set, '
                'by default "%s" is assumed.' % Runner.DISKS_BUS_DEFAULT,
                default=defaultOptions['vmDisksBus'])

        MarketplaceUtil.addEndpointOption(self.parser)

        AuthnCommand.addCloudEndpointOptions(self.parser)

        super(Runnable, self).parse()

        options, self.args = self.parser.parse_args()

        self._assignOptions(defaultOptions, options)


    def _assignOptions(self, defaultOptions, options):
        obj = lambda: None
        Util.assignAttributes(obj, defaultOptions)
        Util.assignAttributes(obj, options.__dict__)
        self.options = obj

    def checkOptions(self):

        if self.options.listType:
            return

        self._checkArgs()

        self.image = self.args[0]

        AuthnCommand.checkCloudEndpointOptionsOnly(self)

        if self.options.extraContextFile and not os.path.isfile(self.options.extraContextFile):
            self.parser.error('Extra context file does not exist')
        if self.options.vncListen and not Util.validateIp(self.options.vncListen):
            self.parser.error('VNC listen IP is not valid')

        MarketplaceUtil.checkEndpointOption(self.options)

        super(Runnable, self).checkOptions()

    def _checkArgs(self):
        if len(self.args) != 1:
            self.parser.error('Please specify the machine image to start')



#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, Centre National de la Recherche Scientifique (CNRS)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
import time
import random
import string

import pika
from pika.adapters import BlockingConnection

class NotificationUseCaseTest(unittest.TestCase):

    queue = 'stratuslab-' + str(random.randint(10000, 99999))

    username = 'guest'
    password = 'guest'

    amqp = {
        'host' : 'dev.rabbitmq.com',
        'port' : 5672,
        'virtual_host' : '/',
        'credentials' : pika.PlainCredentials(username, password)
        }

    machineStates = [ 'CREATED', 'RUNNING', 'DONE' ]

    maxSecondsToWait = 10

    def createMsgRecipients(self):
        option = string.join( 
            [self.amqp['host'],
            self.amqp['virtual_host'],
            self.username,
            self.password,
            self.queue],
            ',')
        return [option]


    def initializeMessageQueue(self):

        # Connect to RabbitMQ
        parameters = pika.ConnectionParameters(**self.amqp)
        connection = BlockingConnection(parameters)
        
        # Open the channel
        channel = connection.channel()
        
        # Declare the queue
        channel.queue_declare(queue=self.queue, 
                              durable=False,
                              exclusive=True, 
                              auto_delete=True)

        print 'Queue: %s' % self.queue

        return (connection, channel)


    def checkNotificationMessages(self, connection, vmId):

        expectedMessages = {}
        for state in self.machineStates:
            message = 'VMID=%i; STATE=%s' % (vmId, state)
            expectedMessages[message] = 1

        channel = connection.channel()

        timedOut = False
        startTime = time.time()
        while len(expectedMessages) > 0 and not timedOut:

            # Call basic get which returns the 3 frame types
            method_frame, header_frame, body = channel.basic_get(queue=self.queue)
            
            # It can be empty if the queue is empty so don't do anything
            if not method_frame.NAME == 'Basic.GetEmpty':
                print "Basic.GetOk %s delivery-tag %i: %s" % (
                    header_frame.content_type,
                    method_frame.delivery_tag,
                    body)
                
                # Acknowledge the receipt of the data
                channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                # Remove message from list of expected messages.
                try:
                    del(expectedMessages[body])
                except KeyError:
                    print 'Unexpected message: %s' % body

            # Don't pound the message server.
            time.sleep(1)                

            # Determine if we've timed out.
            timedOut = (time.time() - startTime) > self.maxSecondsToWait

        if len(expectedMessages) > 0 :
            self.assertEquals(0, len(expectedMessages), 
                              'Unreceived messages:  ' + str(expectedMessages.keys()))


    def cleanUpMessageQueue(self, connection):
        if not connection:
            connection.close()


    def testDummy(self):
        pass


if __name__ == "__main__":
    unittest.main()




#
# Created as part of the StratusLab project (http://stratuslab.eu),
# co-funded by the European Commission under the Grant Agreement
# INFSO-RI-261552."
#
# Copyright (c) 2011, SixSq Sarl
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import shutil
import tempfile

from stratuslab import Util
from stratuslab import Defaults
from stratuslab.ManifestInfo import ManifestInfo
from stratuslab.Compressor import Compressor
from stratuslab.ConfigHolder import ConfigHolder
from stratuslab.Exceptions import InputException

from stratuslab.marketplace.ImageValidator import ImageValidator
from stratuslab.marketplace.ManifestValidator import ManifestValidator
from stratuslab.marketplace.ManifestDownloader import ManifestDownloader

etree = Util.importETree()


class Downloader(object):
    ENDPOINT = Defaults.marketplaceEndpoint
    LOCAL_IMAGE_FILENAME = '/tmp/image.img'

    def __init__(self, configHolder=ConfigHolder()):
        self.localImageFilename = ''
        self.configHolder = configHolder
        self.imageUrl = ''
        self.compression = ''
        self.marketplaceEndpoint = Downloader.ENDPOINT

        configHolder.assign(self)

        self.localImageFilename = os.path.abspath(self.localImageFilename)
        self.manifestObject = None

    def download(self, uri):
        """uri is the full resource uri uniquely identifying a single manifest entry"""
        tempMetadataFilename = tempfile.mktemp()
        ManifestDownloader(self.configHolder).getManifestAsFile(uri, tempMetadataFilename)
        manifestInfo = ManifestInfo(self.configHolder)

        manifestInfo.parseManifestFromFile(tempMetadataFilename)

        tempImageFilename = self._downloadFromLocations(manifestInfo)
        self._verifySignature(tempImageFilename, tempMetadataFilename)

        tempImageFilename = self._inflateImage(tempImageFilename)
        if not os.path.exists(tempImageFilename):
            raise InputException('Failed to find image matching image resource uri: %s' % uri)

        self._verifyHash(tempImageFilename, manifestInfo.sha1)

        shutil.copy2(tempImageFilename, self.localImageFilename)

        os.remove(tempImageFilename)
        os.remove(tempMetadataFilename)

        return self.localImageFilename

    def _downloadFromLocations(self, manifestInfo):
        tempImageFilename = ''
        for location in manifestInfo.locations:
            self._printDetail('Looking for image: %s' % location)
            try:
                tempImageFilename = self._downloadImage(location)
                break
            except KeyboardInterrupt:
                raise
            except:
                pass

        return tempImageFilename

    def _downloadImage(self, url):
        compressionExtension = self._extractCompressionExtension(url)

        localFilename = tempfile.mktemp()
        localImageName = localFilename + compressionExtension

        Util.wget(url, localImageName)

        return localImageName

    def _extractCompressionExtension(self, url):
        compression = url.split('.')[-1]

        if compression in Compressor.compressionFormats:
            return '.' + compression
        else:
            return ''

    def _verifySignature(self, imageFilename, metadataFilename):
        ManifestValidator(self.configHolder).verifySignature(imageFilename, metadataFilename)

    def _inflateImage(self, imageFilename):
        extension = self._extractCompressionExtension(imageFilename)
        inflatedFilename = imageFilename
        if extension:
            self._printDetail('Inflating image %s' % imageFilename)
            Compressor.inflate(imageFilename)
            inflatedFilename = imageFilename[:-len(extension)]

        return inflatedFilename

    def _verifyHash(self, imageFilename, hashFromManifest):
        ImageValidator().verifyHash(imageFilename, hashFromManifest)

    def _printDetail(self, message):
        Util.printDetail(message, self.verboseLevel, Util.VERBOSE_LEVEL_NORMAL)
